"""Interface to ERA5 dataset stored 6 hourly in netCDF formats"""

import os
import numpy as np
from datetime import datetime, timedelta, timezone
from pyproj import Proj
import netCDF4
from NEDAS.grid import Grid
from NEDAS.utils.conversion import units_convert
from NEDAS.datasets import Dataset
from NEDAS.datasets.ecmwf import atmos_utils

class ERA5Data(Dataset):
    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        ##variable dictionary for ERA5 naming convention
        self.variables = {
            'atmos_surf_velocity': {'name':('10U', '10V'), 'is_vector':True, 'units':'m/s'},
            'atmos_surf_temp': {'name':'2T', 'is_vector':False, 'units':'K'},
            'atmos_surf_dewpoint': {'name':'2D', 'is_vector':False, 'units':'K'},
            'atmos_surf_press': {'name':'MSL', 'is_vector':False, 'units':'Pa'},
            'atmos_precip':        {'name':'TP', 'is_vector':False, 'units':'m/s'},
            'atmos_down_longwave': {'name':'STRD', 'is_vector':False, 'units':'W/m2'},
            'atmos_down_shortwave': {'name':'SSRD', 'is_vector':False, 'units':'W/m2'},
            'atmos_surf_vapor_mix': {'getter':self.get_vapmix, 'is_vector':False, 'units':'kg/kg'},
            }

        self.grid = None

    ###format filename
    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        assert kwargs['time'] is not None, 'please specify time'
        assert kwargs['name'] is not None, 'please specify variable name'
        year = '{:04d}'.format(kwargs['time'].year)
        rec = self.variables[kwargs['name']]
        if rec['is_vector']:
            files = []
            for name in rec['name']:
                file = os.path.join(kwargs['path'], f"6h.{name}_{year}.nc")
                assert os.path.exists(file), 'file not found: '+file
                files.append(file)
            return files
        else:
            file = os.path.join(kwargs['path'], f"6h.{rec['name']}_{year}.nc")
            assert os.path.exists(file), 'file not found: '+file
            files = [file]
        return files

    def read_grid(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)[0]
        with netCDF4.Dataset(fname) as f:
            lon = f['longitude'][:].data
            lat = f['latitude'][:].data
            x, y = np.meshgrid(lon, lat)
        self.grid = Grid(Proj('+proj=longlat'), x, y, cyclic_dim='x', pole_dim='y', pole_index=(0,))

    ##find the nearest index in data for the given t
    def find_time_index(self, time_series, time):
        t_ = (time - datetime(1900,1,1,tzinfo=timezone.utc)) / timedelta(hours=1)
        ind = np.abs(time_series - t_).argmin()
        return ind

    def read_var(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        name = kwargs['name']
        assert name is not None, 'please specify which variable (name=?) to get'
        time = kwargs['time']
        k = kwargs.get('k', 0)
        rec = self.variables[name]

        if time is None:
            t_index = 0

        if 'name' in rec:
            files = self.filename(**kwargs)
            if rec['is_vector']:
                var = []
                for i,file in enumerate(files):
                    with netCDF4.Dataset(file) as f:
                        t_in_file = f['time'][:].data
                        t_index = self.find_time_index(t_in_file, time)
                        dat = f.variables[rec['name'][i]][t_index, ...]
                        tmp = dat.data
                        tmp[dat.mask] = np.nan
                        var.append(tmp)
                var = np.array(var)
            else:
                with netCDF4.Dataset(files[0]) as f:
                    t_in_file = f['time'][:].data
                    t_index = self.find_time_index(t_in_file, time)
                    dat = f.variables[rec['name']][t_index, ...]
                    var = dat.data
                    var[dat.mask] = np.nan
            ##convert units if necessary
            if rec['name'] in ('TP', 'STRD', 'SSRD'):
                ##need to convert the fluxes to per second units (they are per 6 hours in the dataset)
                var /= 3600. * 6

        else:
            var = rec['getter'](**kwargs)

        var = units_convert(rec['units'], kwargs['units'], var)
        return var

    def get_vapmix(self, **kwargs):
        dewpoint = self.read_var(**{**kwargs, 'name':'atmos_surf_dewpoint', 'units':'K'})
        e_sat = atmos_utils.satvap(dewpoint)
        press = self.read_var(**{**kwargs, 'name':'atmos_surf_press', 'units':'Pa'})
        vapmix = atmos_utils.vapmix(e_sat, press)
        return vapmix

    def read_obs(self):
        raise NotImplementedError


import importlib
from NEDAS.config import Config
from .base import Inflation

registry = {
    'multiplicative': 'MultiplicativeInflation',
    'RTPP': 'RTPPInflation',
}

def get_inflation_func(c: Config) -> Inflation:
    """
    Get the correct Inflation subclass instance based on configuration

    Args:
        c (Config): Confugration object.

    Returns:
        Inflation: Corresponding Inflation subclass instance.
    """
    if not hasattr(c, 'inflation_def'):
        raise AttributeError("'inflation_def' missing from configuration")
    if not isinstance(c.inflation_def, dict):
        c.inflation_def = {}
    if 'type' not in c.inflation_def.keys():
        raise KeyError("'type' needs to be specified in inflation_def")
    inflation_type = c.inflation_def['type'].split(',')

    prior = ('prior' in inflation_type)
    posterior = ('posterior' in inflation_type)

    adaptive = c.inflation_def.get('adaptive', False)
    coef = c.inflation_def.get('coef', 1.0)

    for key in registry.keys():
        if key in inflation_type:
            module = importlib.import_module('NEDAS.assim_tools.inflation.'+key)
            InflationClass = getattr(module, registry[key])
            return InflationClass(coef, adaptive, prior, posterior)

    raise RuntimeError("No valid inflation class found, check c.inflation_def")



import os
import subprocess
from datetime import timedelta
from NEDAS.config import Config

def generate_truth(c: Config) -> None:
    model = c.models['qg']

    truth_dir = model.truth_dir
    os.system("mkdir -p "+truth_dir)
    print(f"Creating truth run for qg model in {truth_dir}")

    run_dir = os.path.join(truth_dir, 'run')
    init_file = f"output_{c.time_start:%Y%m%d_%H}.bin"
    print(f"Running the model for spinup period to get initial condition: {init_file}")
    opt = {
        'path': run_dir,
        'member': 999,  ##make truth run different from the ensemble members
        'time': c.time_start - model.spinup_hours * timedelta(hours=1),
        'forecast_period': model.spinup_hours,
        'time_start': c.time_start,
        'time_end': c.time_end,
        'debug': c.debug,
        **c.job_submit,
        }
    model.run(**opt)

    c.time = c.time_start
    while c.time < c.time_end:
        opt['time'] = c.time
        opt['forecast_period'] = c.cycle_period

        file = f"output_{c.time:%Y%m%d_%H}.bin"
        next_file = f"output_{c.next_time:%Y%m%d_%H}.bin"
        print(f"Running the model from condition {file} to reach {next_file}")
        model.run(**opt)

        c.time = c.next_time
    print("done.")

    subprocess.run(f"mv -v {run_dir}/*/output*.bin {truth_dir}/.", shell=True, check=True)
    print(f"removing temporary run directory: {run_dir}")
    os.system(f"rm -rf {run_dir}")

if __name__ == '__main__':
    c = Config(parse_args=True)
    generate_truth(c)



import os
import numpy as np
from NEDAS.grid import Grid
from NEDAS.utils.conversion import dt1h
from NEDAS.utils.shell_utils import run_command, run_job, makedir
from NEDAS.models import Model
from .namelist import namelist
from .util import read_data_bin, write_data_bin, grid2spec, spec2grid
from .util import psi2zeta, psi2u, psi2v, psi2temp, uv2zeta, zeta2psi, temp2psi

class QGModel(Model):
    """
    Class for configuring and running the qg model
    """
    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.dx = kwargs['dx'] if 'dx' in kwargs else 1.0
        n = 2*(self.kmax+1)
        self.ny, self.nx = n, n
        x, y = np.meshgrid(np.arange(n), np.arange(n))
        self.grid = Grid(None, x, y, cyclic_dim='xy')
        self.grid.mask = np.full(self.grid.x.shape, False)  ##no grid points are masked

        self.dz = kwargs['dz'] if 'dz' in kwargs else 1.0
        levels = np.arange(0, self.nz, self.dz)

        self.variables = {
            'velocity': {'name':('u', 'v'), 'dtype':'float', 'is_vector':True, 'dt':self.restart_dt, 'levels':levels, 'units':1},
            'streamfunc': {'name':'psi', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':levels, 'units':1},
            'vorticity': {'name':'zeta', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':levels, 'units':1},
            'temperature': {'name':'temp', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':levels, 'units':1},
            }

        self.z_units = 1

        self.run_process = None
        self.run_status = 'pending'

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        if kwargs['member'] is not None:
            mstr = '{:04d}'.format(kwargs['member']+1)
        else:
            mstr = ''

        assert kwargs['time'] is not None, 'missing time in kwargs'
        tstr = kwargs['time'].strftime('%Y%m%d_%H')

        return os.path.join(kwargs['path'], mstr, 'output_'+tstr+'.bin')

    def read_grid(self, **kwargs):
        pass

    def read_mask(self, **kwargs):
        pass

    def read_var(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)

        k = kwargs['k']
        k1 = int(k)
        if k1 < self.nz-1:
            k2 = k1+1
        else:
            k2 = k1

        psik1 = read_data_bin(fname, self.kmax, self.nz, k1)
        psik2 = read_data_bin(fname, self.kmax, self.nz, k2)

        if kwargs['name'] == 'streamfunc':
            var1 = spec2grid(psik1).T
            var2 = spec2grid(psik2).T

        elif kwargs['name'] == 'velocity':
            uk1 = psi2u(psik1)
            vk1 = psi2v(psik1)
            u1 = spec2grid(uk1).T
            v1 = spec2grid(vk1).T
            var1 = np.array([u1, v1])
            uk2 = psi2u(psik2)
            vk2 = psi2v(psik2)
            u2 = spec2grid(uk2).T
            v2 = spec2grid(vk2).T
            var2 = np.array([u2, v2])

        elif kwargs['name'] == 'vorticity':
            zetak1 = psi2zeta(psik1)
            var1 = spec2grid(zetak1).T
            zetak2 = psi2zeta(psik2)
            var2 = spec2grid(zetak2).T

        elif kwargs['name'] == 'temperature':
            tempk1 = psi2temp(psik1)
            var1 = spec2grid(tempk1).T
            tempk2 = psi2temp(psik2)
            var2 = spec2grid(tempk2).T

        ##vertical interp between var1 and var2
        if k1 < self.nz-1:
            return (var1*(k2-k) + var2*(k-k1)) / (k2-k1)
        else:
            return var1

    def write_var(self, var, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)

        k = kwargs['k']
        if k==int(k):
            if kwargs['name'] == 'streamfunc':
                psik = grid2spec(var.T)

            elif kwargs['name'] == 'velocity':
                uk = grid2spec(var[0,...].T)
                vk = grid2spec(var[1,...].T)
                psik = zeta2psi(uv2zeta(uk, vk))

            elif kwargs['name'] == 'vorticity':
                zetak = grid2spec(var.T)
                psik = zeta2psi(zetak)

            elif kwargs['name'] == 'temperature':
                tempk = grid2spec(var.T)
                psik = temp2psi(tempk)

            write_data_bin(fname, psik, self.kmax, self.nz, int(k))

    def z_coords(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        z = np.full(self.grid.x.shape, kwargs['k'])
        return z

    def preprocess(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        restart_dir = kwargs['restart_dir']
        restart_file = self.filename(**{**kwargs, 'path':restart_dir})

        ##restart file to be used in this experiment, in work_dir/cycle/...
        input_file = self.filename(**kwargs)
        input_dir = os.path.dirname(input_file)

        ##just cp the prepared files to the work_dir location
        makedir(input_dir)
        run_command("cp "+restart_file+" "+input_file)

    def postprocess(self, task_id=0, **kwargs):
        pass

    def run(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        task_nproc = self.nproc_per_run
        assert task_nproc==1, f'qg model only support serial runs (got task_nproc={task_nproc})'
        self.run_status = 'running'

        input_file = self.filename(**kwargs)
        run_dir = os.path.dirname(input_file)
        makedir(run_dir)

        time_start = kwargs['time_start']
        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h
        output_file = self.filename(**{**kwargs, 'time':next_time})

        if time >= time_start:
            #this is during cycling
            psi_init_type = 'read'
            prep_input_cmd = 'ln -fs '+input_file+' input.bin; '
        else:
            ##this is initial run for spin up
            psi_init_type = self.psi_init_type
            prep_input_cmd = ''

        qg_exe = os.path.join(self.model_code_dir, 'src', 'qg.exe')

        ##build the shell command line
        shell_cmd = ". "+self.model_env+"; "   ##enter the qg model env
        shell_cmd += "cd "+run_dir+"; "         ##enter the run dir
        shell_cmd += "rm -f restart.nml; "      ##clean up before run
        shell_cmd += prep_input_cmd             ##prepare input file
        shell_cmd += "JOB_EXECUTE "+qg_exe+" . " ##the qg model exe
        shell_cmd += "> run.log 2>&1"               ##output to log

        log_file = os.path.join(run_dir, 'run.log')

        ##give it several tries, each time decreasing time step
        for dt_ratio in [1, 0.6, 0.2]:
            namelist(vars(self), time, forecast_period, psi_init_type, kwargs['member'], dt_ratio, run_dir)

            run_job(shell_cmd, nproc=task_nproc, offset=task_id*task_nproc, **kwargs)

            ##check output
            with open(log_file, 'rt') as f:
                if 'Calculation done' in f.read():
                    break

        ##check output
        with open(log_file, 'rt') as f:
            if 'Calculation done' not in f.read():
                raise RuntimeError('errors in '+log_file)
        if not os.path.exists(os.path.join(run_dir, 'output.bin')):
            raise RuntimeError('output.bin file not found')

        shell_cmd = "cd "+run_dir+"; "
        shell_cmd += "mv output.bin "+output_file
        run_job(shell_cmd, nproc=task_nproc, offset=task_id*task_nproc, **kwargs)




from NEDAS.job_submitters.base import JobSubmitter

class MacOSJobSubmitter(JobSubmitter):

    @property
    def execute_command(self):
        if self.parallel_mode == 'mpi':
            return f"mpirun --map-by :OVERSUBSCRIBE -np {self.nproc}"
        elif self.parallel_mode == 'openmp':
            return f"export OMP_NUM_THREADS={self.nproc};"
        else:
            raise ValueError(f"unknown parallel_mode '{self.parallel_mode}'")




from NEDAS.config import Config
from NEDAS.schemes.get_analysis_scheme import get_analysis_scheme

def main() -> None:
    c = Config(parse_args=True)
    scheme = get_analysis_scheme(c)

    # prepare files
    # initial ensemble
    # truth

    # run analysis scheme
    print("Running NEDAS analysis scheme")
    scheme(c)

if __name__ == '__main__':
    main()



import yaml
import numpy as np
from NEDAS.utils.njit import njit

class EnsembleCovariance:
    def compute_covariance(self,):
        pass

# def covariance(x, y, cov_model='ensemble'):
#     if cov_model == 'ensemble':
#         return None
#     elif cov_model == 'static':
#         return None

@njit
def ensemble_covariance(state_ens, obs_ens):
    """
    Compute ensemble-based covariance between state and observation variables
    Inputs:
    - state_ens: np.ndarray
    State variables, first dimension is ensemble members
    - obs_ens: np.ndarray
    Observation variables, first dimension is ensemble members
    Return:
    - state_variance: np.ndarray, shape is same as state_ens.shape[1:]
    Variance of state variables
    - obs_variance: np.ndarray, shape is same as obs_ens.shape[1:]
    Variance of observation variables
    - corr: np.ndarray, shape is state_ens.shape[1:]+obs_ens.shape[1:]
    Correlation between state and observation variables
    """
    nens = state_ens.shape[0]
    nens_obs = obs_ens.shape[0]
    if nens != nens_obs:
        raise ValueError('Ensemble sizes of state and observation must be equal')
    
    # compute variance of state variables
    state_variance = np.zeros(state_ens.shape[1:])
    for m in range(nens):
        state_variance += state_ens[m,...]**2
    state_variance /= nens-1
    
    # compute variance of observation variables
    obs_variance = np.zeros(obs_ens.shape[1:])
    for m in range(nens):
        obs_variance += obs_ens[m,...]**2
    obs_variance /= nens-1

    # compute covariance between state and observation variables
    new_shape = state_ens.shape[1:]+obs_ens.shape[1:]
    cov = np.zeros(new_shape)
    for m in range(nens):
        cov += np.outer(state_ens[m,...], obs_ens[m,...]).reshape(new_shape)
    cov /= nens-1

    # correlation between state and observation variables
    corr = cov / np.sqrt(np.outer(state_variance, obs_variance).reshape(new_shape))
    
    return state_variance, obs_variance, corr

#def static_covariance(lookup, h_dist, v_dist, t_dist, ):
#
#    return state

def covariance_lookup(c):
    pass



import numpy as np

##nonlinear forward model
def M_nl(x, F, T, dt):
    """
    Lorenz 1996 model with 40 variables, nonlinear advance_time function
    Input:
    -x: np.array, the model state
    -F: parameter, default is 8
    -T: duration of the simulation
    -dt: model time step
    Output:
    -x: np.array, the updated model state after simulation
    """
    for n in range(int(T/dt)):
        x += ((np.roll(x, -1) - np.roll(x, 2)) * np.roll(x, 1) - x + F) * dt
    return x




import numpy as np
import copy

class Grid1D:
    """
    Grid class to handle fields defined on a 1D grid

    Methods convert, interp and distance has the same input args as their counterparts in the Grid class.
    This introduces y coordinates in Grid1D class, although y is not defined for 1D grid, this makes the code that
    calles Grid/Grid1D methods to be easier to maintain.
    """
    def __init__(self, x, bounds=None, regular=True, cyclic=False, distance_type='cartesian', dst_grid=None):

        ##coordinates and properties of the 2D grid
        self.x = x
        self.y = np.zeros(x.size)  ##dummy y coords
        self.regular = regular
        self.cyclic = cyclic
        if bounds is not None:
            self.xmin, self.xmax = bounds
        else:
            self.xmin = np.min(self.x)
            self.xmax = np.max(self.x)
        self.ymin, self.ymax = 0, 0

        self.nx = self.x.size
        self.ny = 1
        if regular:
            self.dx = (self.xmax - self.xmin) / (self.nx - 1)
            self.Lx = self.nx * self.dx
        else:
            self.dx = np.mean(np.diff(np.sort(self.x)))
            self.Lx = self.xmax - self.xmin
        self.Ly = 0
        self.mfx, self.mfy = 1, 1

        if distance_type != 'cartesian':
            raise ValueError(f"distance_type {distance_type} not supported for 1D grid.")

        self._dst_grid = None
        if dst_grid is not None:
            self.set_destination_grid(dst_grid)

    def __eq__(self, other):
        if not isinstance(other, Grid1D):
            return False
        if self.x.size != other.x.size:
            return False
        if not np.allclose(self.x, other.x):
            return False
        return True

    @classmethod
    def regular_grid(cls, xstart, xend, dx, centered=False, **kwargs):
        self = cls.__new__(cls)
        x = np.arange(xstart, xend, dx)
        if centered:
            x += 0.5*dx  ##move coords to center of grid box
        self.__init__(x, regular=True, **kwargs)
        return self

    @classmethod
    def random_grid(cls, xstart, xend, npoints, **kwargs):
        self = cls.__new__(cls)
        x = np.random.uniform(0, 1, npoints) * (xend - xstart) + xstart
        self.__init__(x, bounds=[xstart, xend], regular=False, **kwargs)
        return self

    def change_resolution_level(self, nlevel):
        if not self.regular:
            raise NotImplementedError("change_resolution only works for regular grid now")
        if nlevel == 0:
            return self
        else:
            new_grid = copy.deepcopy(self)
            fac = 2**nlevel
            new_grid.dx = self.dx * fac
            new_grid.nx = int(np.round(self.Lx / new_grid.dx))
            assert new_grid.nx > 1, "Grid.change_resolution_level: new resolution too low, try smaller nlevel"
            new_grid.x = self.xmin + np.arange(new_grid.nx) * new_grid.dx
            return new_grid

    @property
    def dst_grid(self):
        return self._dst_grid

    @dst_grid.setter
    def dst_grid(self, grid):
        assert isinstance(grid, Grid1D), "dst_grid should be a Grid1D instance"
        if grid == self.dst_grid:  ##the same grid is set before
            return
        self._dst_grid = grid

        ##prepare indices and weights for interpolation
        ##when dst_grid is set, these info are prepared and stored to avoid recalculating
        ##too many times, when applying the same interp to a lot of flds
        inside, vertices, in_coords, nearest = self.find_index(grid.x)
        self.interp_inside = inside
        self.interp_vertices = vertices
        self.interp_nearest = nearest
        self.interp_weights = self._interp_weights(inside, vertices, in_coords)

        ##prepare indices for coarse-graining
        inside, _, _, nearest = self.dst_grid.find_index(self.x)
        self.coarsen_inside = inside
        self.coarsen_nearest = nearest

    def set_destination_grid(self, grid):
        self.dst_grid = grid

    def wrap_cyclic(self, x_):
        if self.cyclic:
            x_ = np.mod(x_ - self.xmin, self.Lx) + self.xmin
        return x_

    def find_index(self, x_):
        x_ = np.array(x_).flatten()
        x_ = self.wrap_cyclic(x_)
        xi = self.x
        i_ = np.argsort(xi)
        xi_ = xi[i_]
        ##pad cyclic coordinates with additional grid point
        if self.cyclic:
            if xi_[0]+self.Lx not in xi_:
                xi_ = np.hstack((xi_, xi_[0] + self.Lx))
                i_ = np.hstack((i_, i_[0]))

        pi = np.array(np.searchsorted(xi_, x_, side='right'))
        inside = ~np.logical_or(pi==len(xi_), pi==0)
        pi = pi[inside]

        vertices = np.zeros(pi.shape+(2,), dtype=int)
        vertices[:, 0] = i_[pi-1]
        vertices[:, 1] = i_[pi]
        in_coords = (x_[inside] - xi_[pi-1]) / (xi_[pi] - xi_[pi-1])

        nearest = np.zeros(pi.shape, dtype=int)
        ind = np.where(in_coords<0.5)
        nearest[ind] = i_[pi-1][ind]
        ind = np.where(in_coords>=0.5)
        nearest[ind] = i_[pi][ind]

        return inside, vertices, in_coords, nearest

    def _interp_weights(self, inside, vertices, in_coords):
        ##compute bilinear interp weights
        interp_weights = np.zeros(vertices.shape)
        interp_weights[:, 0] =  1-in_coords
        interp_weights[:, 1] =  in_coords
        return interp_weights

    def interp(self, fld, x=None, y=None, method='linear'):
        if x is None:
            ##use precalculated weights for self.dst_grid
            inside = self.interp_inside
            vertices = self.interp_vertices
            nearest = self.interp_nearest
            weights = self.interp_weights
            x = self.dst_grid.x
        else:
            ##otherwise compute the weights for the given x,y
            inside, vertices, in_coords, nearest = self.find_index(x)
            weights = self._interp_weights(inside, vertices, in_coords)

        fld_interp = np.full(np.array(x).flatten().shape, np.nan)
        if fld.shape == self.x.shape:
            if method == 'nearest':
                # find the node of the triangle with the maximum weight
                fld_interp[inside] = fld.flatten()[nearest]
            elif method == 'linear':
                # sum over the weights for each node of triangle
                fld_interp[inside] = np.einsum('nj,nj->n', np.take(fld.flatten(), vertices), weights)
            else:
                raise ValueError("'method' should be 'nearest' or 'linear'")
        else:
            raise ValueError("field shape does not match grid shape, or number of triangle elements")
        return fld_interp.reshape(np.array(x).shape)

    ###utility function for coarse-graining (high->low resolution)
    def coarsen(self, fld):
        ##find which location x_,y_ falls in in dst_grid
        if fld.shape == self.x.shape:
            inside = self.coarsen_inside
            nearest = self.coarsen_nearest
        else:
            raise ValueError("field shape does not match grid shape")

        fld_coarse = np.zeros(self.dst_grid.x.flatten().shape)
        count = np.zeros(self.dst_grid.x.flatten().shape)
        fld_inside = fld.flatten()[inside]
        valid = ~np.isnan(fld_inside)  ##filter out nan

        ##average the fld points inside each dst_grid box/element
        np.add.at(fld_coarse, nearest[valid], fld_inside[valid])
        np.add.at(count, nearest[valid], 1)

        valid = (count>1)  ##do not coarse grain if only one point near by
        fld_coarse[valid] /= count[valid]
        fld_coarse[~valid] = np.nan

        return fld_coarse.reshape(self.dst_grid.x.shape)

    def convert(self, fld, is_vector=False, method='linear', coarse_grain=False, **kwargs):
        if self.dst_grid != self:
            fld_out = np.full(self.dst_grid.x.shape, np.nan)
            fld_out = self.interp(fld, method=method)
            if coarse_grain:
                ##coarse-graining if more points fall in one grid
                fld_coarse = self.coarsen(fld)
                ind = ~np.isnan(fld_coarse)
                fld_out[ind] = fld_coarse[ind]
        else:
            fld_out = fld
        return fld_out

    def distance(self, ref_x, x, ref_y=None, y=None, p=1):
        ##normal cartesian distances in x and y
        dist_x = np.abs(x - ref_x)
        ##if there are cyclic boundary condition, take care of the wrap around
        if self.cyclic:
            dist_x = np.minimum(dist_x, self.Lx - dist_x)
        return dist_x



import os
import inspect
from typing import Optional, Any
import yaml
import dateutil.parser
import numpy as np
from datetime import datetime, timedelta, timezone
from pyproj import Proj
import NEDAS
from .parse_config import parse_config
from NEDAS.utils import parallel

class Config:
    """
    Class to manage the configuration for running the NEDAS analysis.
    Configuration entries are described in details in :doc:`config_file`.

    Args:
        config_file (str, optional): Path to the configuration file.
        parse_args (bool, optional): If true, parse command line arguments to collect configuration. Default is False.
        **kwargs: Additional key-value pairs to be passed to parse_config. Can be used to override values in the config file.

    Attributes:
        nproc (int): Number of processors to use for the analysis step.
        comm (Comm): MPI communicator, set by :meth:`set_comm`.
        grid (Grid): Analysis grid, set by :meth:`set_analysis_grid`.
        model_config (dict): A dictionary where keys are model names and values are the corresponding Model instances, set by :meth:`set_model_config`.
        dataset_config (dict): A dictionary where keys are dataset names and values are the corresponding Dataset instances, set by :meth:`set_dataset_config`.
    """
    work_dir: str
    time: datetime
    time_start: datetime
    time_end: datetime
    time_analysis_start: datetime
    time_analysis_end: datetime
    cycle_period: float
    nproc: int
    nproc_mem: int
    nproc_rec: int
    nproc_util: int
    pid: int
    pid_mem: int
    pid_rec: int
    grid_def: dict
    model_def: dict
    dataset_def: dict
    comm: parallel.Comm

    def __init__(self, config_file: Optional[str]=None, parse_args: bool=False, **kwargs):
        # parse the yaml config file to obtain the values
        code_dir = os.path.dirname(inspect.getfile(self.__class__))
        self.config_dict = parse_config(code_dir, config_file, parse_args, **kwargs)

        # replace placeholders in dir paths with actual values
        self.work_dir = os.path.abspath(self.work_dir)
        self.nedas_root = NEDAS.__path__[0]
        self.config_dict = self.parse_directories(self.config_dict)

        # set additional attributes
        self.set_time()
        self.set_comm()
        self.set_analysis_grid()
        self.set_models()
        self.set_datasets()

        if 'pid_show' not in self.config_dict or self.config_dict['pid_show'] is None:
            self.config_dict['pid_show'] = 0

        self.directories: dict[str, str] = self.config_dict['directories']

    def __getattr__(self, key):
        # get values from config_dict if defined, otherwise will get the attr from the instance directly.
        if 'config_dict' in self.__dict__ and key in self.config_dict:
            return self.config_dict[key]

    def __setattr__(self, key, value):
        if key != 'config_dict' and 'config_dict' in self.__dict__ and key in self.config_dict:
            if key == 'time' and not isinstance(value, datetime):
                raise TypeError("Time must be a datetime object.")
            self.config_dict[key] = value
        else:
            super().__setattr__(key, value)

    @property
    def prev_time(self) -> datetime:
        """
        Previous analysis time. Automatically updated when self.time changes.

        Returns:
            datetime: Previous analysis time.
        """
        if self.time > self.time_start:
            return self.time - self.cycle_period * timedelta(hours=1)
        else:
            return self.time

    @property
    def next_time(self) -> datetime:
        """
        Next analysis time. Automatically updated when self.time changes.

        Returns:
            datetime: Next analysis time.
        """
        return self.time + self.cycle_period * timedelta(hours=1)

    @property
    def print_1p(self):
        """
        Customized print function for showing runtime message.

        Only the processor with ID = self.pid_show will show the message,
        this avoids the redundancy if all processors are showing the same message.
        """
        decorator = parallel.by_rank(self.comm, self.pid_show)
        return decorator(parallel.print_with_cache)

    def cycle_dir(self, time: datetime) -> str:
        """
        Directory path for an analysis cycle.

        Args:
            time (datetime): Time of the analysis cycle.

        Returns:
            str: Directory path for the analysis cycle.
        """
        return self.directories['cycle_dir'].format(time=time)

    def forecast_dir(self, time: datetime, model_name: str):
        """
        Directory path for a model forecast step.

        Args:
            time (datetime): Time of the analysis cycle.
            model_name (str): Name of the model.

        Returns:
            str: Directory path for the model forecast.
        """
        return self.directories['forecast_dir'].format(time=time, model_name=model_name)

    def analysis_dir(self, time: datetime, iter: int=0):
        """
        Directory path for an analysis step.

        Args:
            time (datetime): Time of the analysis cycle.
            iter (int): If niter > 1, an outer iteration loop exists, step is the index in the loop.

        Returns:
            str: Directory path for the analysis step.
        """
        if self.niter == 1:
            iter_dir= ''
        else:
            iter_dir = f"iter{iter}"
        return self.directories['analysis_dir'].format(time=time, step=iter_dir)

    def parse_directories(self, data: Any) -> Any:
        """
        Parse the directories or file names defined in :code:`data`
        and replace the placeholders {work_dir} and {nedas_root} with the actual values.
        """
        if isinstance(data, dict):
            return {key: self.parse_directories(value) for key, value in data.items()}
        elif isinstance(data, list):
            return [self.parse_directories(element) for element in data]
        elif isinstance(data, str):
            return data.replace('{work_dir}', self.work_dir).replace('{nedas_root}', self.nedas_root)
        else:
            return data

    def set_time(self):
        """
        Initialize the time variables for the analysis.

        Checks if the mandatory :code:`time_*` entries are defined in the config file.
        If :code:`time` is not set, set it to :code:`time_start` by default.
        YAML file recognizes 2001-01-01T00:00:00 format and convert directly to datetime object.
        If time is a formatted string, will try to parse it using dateutil.parser.
        """
        # check if mandatory time keys are defined in config file
        for key in ['time', 'time_start', 'time_end', 'time_analysis_start', 'time_analysis_end']:
            if key not in self.config_dict:
                raise KeyError(f"'{key}' is missing in config file")
            if isinstance(self.config_dict[key], str):
                try:
                    self.config_dict[key] = dateutil.parser.parse(self.config_dict[key])
                except Exception:
                    raise ValueError(f"Failed to convert string {key}={self.config_dict[key]} to datetime")
            # add default tzinfo
            if self.config_dict[key] and self.config_dict[key].tzinfo is None:
                self.config_dict[key] = self.config_dict[key].replace(tzinfo=timezone.utc)

        if self.time is None:
            ##initialize current time to start time, if not available
            self.config_dict['time'] = self.config_dict['time_start'].replace()

    def set_comm(self):
        """
        Initialize the MPI communicator, check the number of processors, split the communicator if necessary.

        For serial program, use a dummy communicator, set :code:`nproc` to the number of available processors on the machine;
        for MPI program, use :code:`MPI.COMM_WORLD` and check if size matchs with :code:`nproc`.

        Split the communicator into member and record groups, according to :code:`nproc` and :code:`nproc_mem`.
        See :mod:`NEDAS.utils.parallel` module for more details.
        """
        ##initialize mpi communicator (could be size 1 for serial program)
        self.comm = parallel.Comm()
        comm_size = self.comm.Get_size()
        self.pid = self.comm.Get_rank()  ##current processor id

        ##self.nproc is the total number of processors
        if hasattr(self, 'nproc') and self.nproc is not None:
            ##if it is set in config file, check if it matches the actual comm size
            if comm_size > 1 and self.nproc != comm_size:
                raise ValueError(f"Config nproc={self.nproc} does not match MPI comm size={comm_size}.")
        else:
            ##if not set, figure out the available number of processors
            if comm_size == 1:
                ##serial program, set nproc to available processors (scheduler can spawn tasks to them)
                nproc_avail = os.cpu_count()
                if nproc_avail:
                    self.nproc = nproc_avail
            else:
                ##mpi program, set nproc to comm size
                self.nproc = comm_size

        ##divide processors into mem/rec groups
        if not hasattr(self, 'nproc_mem') or self.nproc_mem is None:
            self.nproc_mem = self.nproc
        if self.nproc % self.nproc_mem != 0:
            raise ValueError(f"nproc={self.nproc} is not evenly divided by nproc_mem={self.nproc_mem}")
        self.nproc_rec = int(self.nproc/self.nproc_mem)
        self.pid_mem = self.pid % self.nproc_mem
        self.pid_rec = self.pid // self.nproc_mem

        if not hasattr(self, 'nproc_util') or self.nproc_util is None:
            self.nproc_util = self.nproc

        ##split comm if in mpi program
        if comm_size > 1:
            self.comm_mem = self.comm.Split(self.pid_rec, self.pid_mem)
            self.comm_rec = self.comm.Split(self.pid_mem, self.pid_rec)
        else:
            self.comm_mem = self.comm
            self.comm_rec = self.comm

    def set_analysis_grid(self):
        """
        Initialize the analysis grid based on the configuration.

        If :code:`grid_def['type']` is 'custom', will create a analysis grid based on provided parameters.
        If :code:`grid_def['type']` is a model name, will load the grid from the specified model class.
        """
        assert isinstance(self.grid_def, dict)
        if self.grid_def['type'] == 'custom':
            if 'proj' in self.grid_def and self.grid_def['proj'] is not None:
                proj = Proj(self.grid_def['proj'])
            else:
                proj = None
            xmin, xmax = self.grid_def['xmin'], self.grid_def['xmax']
            ymin, ymax = self.grid_def['ymin'], self.grid_def['ymax']
            dx = self.grid_def['dx']
            known_keys = {'type', 'proj', 'xmin', 'xmax', 'ymin', 'ymax', 'dx', 'mask'}
            other_opts = {k: v for k, v in self.grid_def.items() if k not in known_keys}
            self.grid = NEDAS.grid.Grid.regular_grid(proj, xmin, xmax, ymin, ymax, dx, **other_opts)

            ##mask for invalid grid points (none for now, add option later)
            self.grid.mask = np.full((self.grid.ny, self.grid.nx), False, dtype=bool)
            if 'mask' in self.grid_def and self.grid_def['mask'] is not None:
                model_name = self.grid_def['mask']
                Model = NEDAS.models.get_model_class(model_name)
                model = Model()
                if hasattr(model, 'prepare_mask'):
                    self.grid.mask = model.prepare_mask(self.grid)

        else:
            ##get analysis grid from model module
            model_name = self.grid_def['type']
            kwargs = self.model_def[model_name]
            Model = NEDAS.models.get_model_class(model_name)
            model = Model(**kwargs)
            self.grid = model.grid

    def set_models(self):
        """
        Initialize model instances based on :code:`model_def[model_name]` settings.
        Store the model instances in :code:`models[model_name]`.
        """
        self.models = {}
        for model_name, kwargs in self.model_def.items():
            Model = NEDAS.models.get_model_class(model_name)
            if not isinstance(kwargs, dict):
                kwargs = {}
            self.models[model_name] = Model(**kwargs)

    def set_datasets(self):
        """
        Initialize dataset instances based on :code:`dataset_def[dataset_name]` settings.
        Store the dataset instances in :code:`datasets[dataset_name]`.
        """
        self.datasets = {}
        for dataset_name, kwargs in self.dataset_def.items():
            Dataset = NEDAS.datasets.get_dataset_class(dataset_name)
            if not isinstance(kwargs, dict):
                kwargs = {}
            self.datasets[dataset_name] = Dataset(grid=self.grid, mask=self.grid.mask, **kwargs)

    def show_summary(self):
        """
        Print a summary of the configuration.
        """
        print(f"""Initializing config...
 working directory: {self.work_dir}
 parallel scheme: nproc = {self.nproc}, nproc_mem = {self.nproc_mem}
 cycling from {self.time_start} to {self.time_end}
 analysis start at {self.time_analysis_start}
 cycle_period = {self.cycle_period} hours
 current time: {self.time}
 """, flush=True)

    def dump_yaml(self, config_file: str):
        """
        Dump the current configuration to a YAML file.

        Args:
            config_file (str): Path to the output configuration file.
        """
        with open(config_file, 'w') as f:
            yaml.dump(self.config_dict, f)



import os
from functools import wraps
import time
from concurrent.futures import ProcessPoolExecutor
import threading
import traceback
import numpy as np
from NEDAS.utils.progress import print_with_cache, progress_bar

def check_parallel_io() -> bool:
    """
    Check if netCDF4 is built with parallel I/O support.

    Returns:
        bool: True if netCDF4 module support parallel I/O mode.
    """
    try:
        from netCDF4 import Dataset
        with Dataset('dummy.nc', mode='w', parallel=True):
            return True
    except Exception:
        return False

class Comm:
    """
    Communicator class supporting both serial and MPI programs.

    When the python program is started with MPI environment, for example::

    $ mpirun -n 10 python -m mpi4py program.py

    A communicator can be obtained from the mpi4py package:

    >>> from mpi4py import MPI
    >>> comm = MPI.COMM_WORLD

    However, when the program is run in 

    Attributes:
        parallel_io (bool): If netCDF4.Dataset is built with parallel I/O support.

    """

    def __init__(self):
        ##detect if mpi environment exists
        ##possible environ variable names from mpi calls
        mpi_env_var = ('PMI_SIZE', 'OMPI_UNIVERSE_SIZE')
        if any([ev in os.environ for ev in mpi_env_var]):
            ##program is called from mpi, initialize comm
            try:
                from mpi4py import MPI
                self._MPI = MPI
                self._comm = MPI.COMM_WORLD

            except ImportError:
                print("Warning: MPI environment found but 'mpi4py' module is not installed. Falling back to serial program for now.", flush=True)
                self._MPI = None
                self._comm = DummyComm()

        else:
            ##serial program, use a dummy communicator
            self._MPI = None
            self._comm = DummyComm()

        self.parallel_io = check_parallel_io()

        ##file lock to ensure only one processor access a file at a time
        self._locks = {}

    def __getattr__(self, attr):
        if hasattr(self._comm, attr):
            return getattr(self._comm, attr)
        raise AttributeError

    def init_file_lock(self, filename):
        """
        Initialize file locks for thread-safe I/O.

        Args:
            filename (str): Path to the file.
        """
        if self._MPI is None:
            return
        if isinstance(self._comm, DummyComm):
            return
        if not filename:
            return
        if filename not in self._locks:
            ##create the lock memory
            if self.Get_rank() == 0:
                lock_mem = np.zeros(1, dtype='B')
            else:
                lock_mem = None
            lock_win = self._MPI.Win.Create(lock_mem, comm=self._comm)
            self._locks[filename] = lock_win

    def cleanup_file_locks(self):
        for file, lock_win in self._locks.items():
            lock_win.Free()

    def acquire_file_lock(self, filename):
        if self._MPI is None:
            return
        if isinstance(self._comm, DummyComm):
            return
        assert filename in self._locks, f"Comm: file lock for {filename} not initialized"
        lock_win = self._locks[filename]
        check_dt = 0.1  ##check file locks every 0.1 seconds, can make this configurable
        while True:
            # print(f"pid {self.Get_rank()} waiting for lock on {filename}", flush=True)
            lock_mem = np.zeros(1, dtype='B')
            one = np.array([1], dtype='B')
            lock_win.Lock(0, self._MPI.LOCK_EXCLUSIVE)
            lock_win.Fetch_and_op(one, lock_mem, 0, 0, self._MPI.REPLACE)
            lock_win.Unlock(0)
            if lock_mem[0] == 0:
                # print(f"pid {self.Get_rank()} acquires lock on {filename}", flush=True)
                break
            time.sleep(check_dt)

    def release_file_lock(self, filename):
        if self._MPI is None:
            return
        if isinstance(self._comm, DummyComm):
            return
        if filename in self._locks:
            zero = np.array([0], dtype='B')
            lock_win = self._locks[filename]
            lock_win.Lock(0, self._MPI.LOCK_EXCLUSIVE)
            lock_win.Put(zero, 0, 0)
            lock_win.Unlock(0)
            # print(f"pid {self.Get_rank()} releases lock on {filename}", flush=True)

    def finalize(self):
        if self._MPI is not None:
            self._MPI.Finalize()

class DummyComm:
    """Dummy communicator for python without mpi"""
    def __init__(self):
        self.size = 1
        self.rank = 0
        self.buf = {}

    def Get_size(self):
        return self.size

    def Get_rank(self):
        return self.rank

    def Barrier(self):
        pass

    def Split(self, color=0, key=0):
        return self

    def bcast(self, obj, root=0):
        return obj

    def send(self, obj, dest, tag):
        self.buf[tag] = obj

    def recv(self, source, tag):
        return self.buf[tag]

    def allgather(self, obj):
        return [obj]

    def gather(self, obj, root=0):
        return obj

    def allreduce(self, obj):
        return obj

    def reduce(self, obj, root=0):
        return obj

def by_rank(comm, rank):
    """
    Decorator for func() to be run only by rank 0 in comm
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if comm.Get_rank() == rank:
                result = func(*args, **kwargs)
            else:
                result = None
            return result
        return wrapper
    return decorator

def bcast_by_root(comm):
    """
    Decorator for func() to be run only by rank 0 in comm,
    and result of func() is then broadcasted to all other ranks.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if comm.Get_rank() == 0:
                result = func(*args, **kwargs)
            else:
                result = None
            result = comm.bcast(result, root=0)
            return result
        return wrapper
    return decorator

def distribute_tasks(comm, tasks, load=None):
    """
    Divide a list of task indices and assign a subset to each rank in comm

    Args:
        comm (Comm): MPI communicator
        tasks (list): List of task indices (to be distributed over the processors)
        load (np.ndarray, optional):
            Amount of workload for each task element
            The default is None, we will let tasks have equal workload

    Returns:
        dict: Dictionary {rank:list}, list is the subset of tasks for the processor rank
            calling this function to work on
    """
    nproc = comm.Get_size()  ##number of processors
    ntask = len(tasks)       ##number of tasks

    ##assume equal load between tasks if not specified
    if load is None:
        load = np.ones(ntask)

    assert load.size==ntask, f'load.size = {load.size} not equal to tasks.size'

    ##normalize to get load distribution function
    load = load / np.sum(load)

    ##cumulative load distribution, rounded to 5 decimals
    cum_load = np.round(np.cumsum(load), decimals=5)

    ##given the load distribution function, we assign load to processors
    ##by evenly divide the distribution into nproc parts
    ##this is done by searching for r/nproc in the cumulative load for rank r
    ##task_id holds the start/end index of task for each rank in a sequence
    task_id = np.zeros(nproc+1, dtype=int)

    target_cum_load = np.arange(nproc)/nproc  ##we want even distribution of load
    tol = 0.1/nproc  ##allow some tolerance for rounding error in comparing cum_load to target_cum_load
    ind1 = np.searchsorted(cum_load+tol, target_cum_load, side='right')
    ind2 = np.searchsorted(cum_load-tol, target_cum_load, side='right')

    ##choose between ind1,ind2, whoever gives best match between cum_load[ind?] and target_cum_load
    task_id[0:-1] = np.where(np.abs(cum_load[ind1-1]-target_cum_load) < np.abs(cum_load[ind2-1]-target_cum_load), ind1, ind2)

    ##make sure the two end points are right
    task_id[0] = 0
    task_id[-1] = ntask

    ##dict for each rank r -> its own task list given start/end index
    task_list = {}
    for r in range(nproc):
        task_list[r] = tasks[task_id[r]:task_id[r+1]]

    return task_list

class Scheduler:
    """
    A scheduler class for queuing and running multiple jobs on available workers (group of processors).
    The jobs are submitted by one processor with the scheduler, while the job.run code is calling subprocess
    to be run on the worker
    """
    def __init__(self, nworker, walltime=None, check_dt=0.1, debug=False):
        self.nworker = nworker
        self.available_workers = list(range(nworker))
        self.walltime = walltime
        self.check_dt = check_dt
        self.debug = debug
        self.jobs = {}
        self.executor = ProcessPoolExecutor(max_workers=nworker)
        self.queue_open = True
        self.running_jobs = []
        self.pending_jobs = []
        self.completed_jobs = []
        self.error_jobs = {}
        self.njob = 0

    def submit_job(self, name, job, *args, **kwargs):
        """
        Submit a job to the scheduler, hold info in jobs dict
        Input:
        - name is a unique name (str) to identify this job
        - job is an object with run, is_running and kill methods
        - args,kwargs are to be passed into job.run()
        """
        self.jobs[name] = {'worker_id':None, 'start_time':None, 'job':job,
                           'args': args, 'kwargs': kwargs, 'future':None }
        self.pending_jobs.append(name)
        self.njob += 1
        if self.debug:
            print(f"Scheduler: Job {name} added: '{job.__name__, args, kwargs}'", flush=True)

    def monitor_job_queue(self):
        """
        Monitor the available_workers and pending_jobs, assign a job to a worker if possible
        Monitor the running_jobs for jobs that are finished, kill jobs that exceed walltime,
        and move the finished jobs to completed_jobs
        """
        while len(self.completed_jobs) < self.njob:

            ##assign pending job to available workers
            while self.available_workers and self.pending_jobs and self.queue_open:
                worker_id = self.available_workers.pop(0)
                name = self.pending_jobs.pop(0)
                info = self.jobs[name]
                info['worker_id'] = worker_id
                info['start_time'] = time.time()
                info['future'] = self.executor.submit(info['job'], worker_id, *info['args'], **info['kwargs'])
                self.running_jobs.append(name)
                if self.debug:
                    print(f"Scheduler: Job {name} started by worker {worker_id}", flush=True)

            ##if there are completed jobs, free up their workers
            names = [name for name in self.running_jobs if self.jobs[name]['future'].done()]
            for name in names:
                ##catch errors from job
                try:
                    self.jobs[name]['future'].result()
                except Exception as e:
                    tb = traceback.format_exc()
                    print(f'Scheduler: Job {name} raised exception: \n{tb}', flush=True)
                    self.error_jobs[name] = tb
                    #return  ###if exit right away and don't wait for other jobs to finish, uncomment this
                self.running_jobs.remove(name)
                self.completed_jobs.append(name)
                self.available_workers.append(self.jobs[name]['worker_id'])
                if self.debug:
                    print(f"Scheduler: Job {name} completed", flush=True)

            ##kill jobs that exceed walltime
            if self.walltime is not None:
                for name in self.running_jobs:
                    elapsed_time = time.time() - self.jobs[name]['start_time']
                    if elapsed_time > self.walltime:
                        self.jobs[name]['future'].cancel()
                        self.running_jobs.remove(name)
                        self.available_workers.append(self.jobs[name]['worker_id'])
                        e = RuntimeError(f'Scheduler: Job {name} exceeds walltime ({self.walltime}s)')
                        self.error_jobs[name] = e
                        self.completed_jobs.append(name)

            ##just show a progress bar if not output debug messages
            if not self.debug:
                print_with_cache(progress_bar(len(self.completed_jobs), self.njob+1))

            time.sleep(self.check_dt)

    def start_queue(self):
        """
        Start the job queue, and wait for jobs to complete
        """
        try:
            monitor_thread = threading.Thread(target=self.monitor_job_queue)
            monitor_thread.start()
            monitor_thread.join()
        except KeyboardInterrupt:
            self.queue_open = False
            ##kill running jobs
            self.shutdown()

    def shutdown(self):
        if self.error_jobs:
            error_details = "\n".join([f"ERROR: Job {job}: {error}" for job, error in self.error_jobs.items()])
            raise RuntimeError(f'Scheduler: there are jobs with errors:\n{error_details}')
        self.executor.shutdown(wait=True)



###adapted from NERSC-HYCOM-CICE/pythonlibs/abfile
##Module for doing IO on files used by hycom
import numpy
import struct
import sys
import re

# Firldnames as they appear in ordered form in the regional grid files
grid_ordered_fieldnames = [
    "plon", "plat", "qlon", "qlat", "ulon", "ulat", "vlon", "vlat", "pang", "scpx",
    "scpy", "scqx", "scqy", "scux", "scuy", "scvx", "scvy", "cori", "pasp"
    ]

class AFileError(Exception) :
    pass

class BFileError(Exception) :
    pass

class AFile :
    """ Class for doing binary input/output on hycom .a files. Normally used by 
    ABFile* classes, but can be called by itself to read a .a-file """
    huge = 2.0**100
    def __init__(self,idm,jdm,filename,action,mask=False,real4=True,endian="big") :
        self._idm = idm
        self._jdm = jdm
        self._filename = filename
        self._action = action
        self._mask = mask
        self._real4= real4
        self._endian= endian
        if self._action.lower() not in ["r","w","r+"]  :
            raise AFileError("action argument must be either r(ead) or w(rite) or r+")
        if self._endian.lower() not in ["little","big","native"]  :
            raise AFileError("action argument must be either native, little or big")

        if self._endian.lower() == "native" :
            self._endian=sys.byteorder

        if self._endian.lower() == "big" :
            self._endian_structfmt = ">"
        else :
            self._endian_structfmt = "<"

        self._init_record()
        self._open()

    def _init_record(self) :
        # Size of output 2D Array
        self._n2drec= ((self._idm*self._jdm+4095)//4096)*4096 # original

        # Init sequenctial record counter
        self._iarec = 0
        self._spval = 2**100.

    def _open(self) :
        # Open .a and .b file
        self._filea = open(self._filename,self._action+"b")

    def writerecord(self,h,mask=None,record=None) :
        """ Write one record to file.

        Arguments:    
          h      : data field to write. Must conform to shape (self._jdm, self._idm)
          mask  : where mask is set to True, h is set to spval. Ignored if masking not set in class

        Keyword Arguments:
          record:  not implemented yet
        """

        # Initialize writing array (1D)
        w=numpy.ones(self._n2drec) *self._spval

        # Check array shape against idm/jdm
        if h.shape[0] != self._jdm or h.shape[1] != self._idm :
            raise AFileError("array shape is (%d,%d),expected (%d,%d)"%(h.shape[0],h.shape[1],self._jdm,self._idm))

        # Fill w array
        w[0:self._idm*self._jdm] = h.flatten()

        # Seek if provided, otherwise use current position
        if record is not None : self.seekrecord(record)

        # Calc min and mask
        #print("mask boolean:",self._mask)
        if self._mask :
            if mask is None:
                mask = numpy.isnan(h)
            I=numpy.where(~mask)
            hmax=h[I].max()
            hmin=h[I].min()
            J=numpy.where(mask.flatten())
            w[J] = self._spval
          # print "writerecord w mask:",numpy.count_nonzero(mask),mask.size
          # print "writerecord w mask:",hmin,hmax,w[0:self._idm*self._jdm].min(),w[0:self._idm*self._jdm].max()
        else :
            hmax=h.max()
            hmin=h.min()
            #print "writerecord wo mask:",hmin,hmax

        if self._real4 :
            struct_fmt="f"
        else :
            struct_fmt="d"

        # 1) Use struct
        #binpack=struct.pack("%s%d%s"%(self._endian_structfmt,w.size,struct_fmt),*w[:])
        #self._filea.write(binpack)

        # 2) Alternative using numpy.ndarray.tofile
        w = w.astype(numpy.dtype("%s%s"%(self._endian_structfmt,struct_fmt)))
        #print w.dtype,w.dtype.byteorder
        w.tofile(self._filea)




        return hmin,hmax

    def read_record(self,record) :
        self.seekrecord(record)

        ##1) Use struct
        #if self._real4 :
        #    raw = self._filea.read(self.n2drec*4)
        #    fmt =  "%s%df"%(self._endian_structfmt,self.n2drec)
        #else :
        #    raw = self._filea.read(self.n2drec*8)
        #    fmt =  "%s%dd"%(self._endian_structfmt,self.n2drec)
        #w =  numpy.array(struct.unpack(fmt,raw))

        #2) Use numpy fromfile
        if self._real4 :
            struct_fmt="f"
        else :
            struct_fmt="d"
        mydtype=numpy.dtype("%s%s"%(self._endian_structfmt,struct_fmt))
        w=numpy.fromfile(self._filea,dtype=mydtype,count=int(self.n2drec))

        w=w[0:self.idm*self.jdm]
        w.shape=(self.jdm,self.idm)
        #print w.min(),w.max()
        w=numpy.ma.masked_where(w>self.huge*.5,w)
        #print w.min(),w.max()


        return w

    def seekrecord(self,record) :
        # Seek to correct record and read
        if self._real4 :
            self._filea.seek(int(record*self.n2drec*4))
        else :
            self._filea.seek(int(record*self.n2drec*8))
        return

    def close(self) :
        self._filea.close()

    @property
    def n2drec(self):
        return self._n2drec

    @property
    def idm(self):
        return self._idm

    @property
    def jdm(self):
        return self._jdm

class ABFile:
    """ Class for doing input/output on pairs of hycom .a and .b files. Base class, 
    not meant to be used directly """

    def __init__(self,basename,action,mask=False,real4=True,endian="big") :
        #self._basename=basename
        self._basename=ABFile.strip_ab_ending(basename) # Ensure .ab ending is stripped
        #print basename,self._basename
        self._action=action
        self._fileb = open(self._basename+".b",self._action)
        self._filea = None
        self._mask = mask
        self._real4 = real4
        self._endian = endian
        self._firstwrite=True

    def scanitem(self,item=None,conversion=None) :
        line = self._fileb.readline().strip()
        if item is not None :
            pattern=r"^(.*)'(%-6s)'[ =]*"%item
            m=re.match(pattern,line)
        else :
            m=re.match(r"^(.*)'(.*)'[ =]*",line)
        if m :
            if conversion :
                value = conversion(m.group(1))
                return m.group(2),value
            return m.group(2), m.group(1)
        else :
            return None,None

    def writeitem(self,key,value) :
        if value is None :
            msg = "key %s is not defined "%key
            raise ValueError(msg)
        if type(value) == type(1) :
            tmp ="%5d    '%-6s'\n"%(value,key)
        else :
            msg = "writeitem not implemented for this type: %s"%type(value)
            raise NotImplementedError(msg)
        self._fileb.write(tmp)

    def readline(self) :
        return self._fileb.readline()

    def bminmax(self,*arks,**kwargs) :
        raise BFileError("bminmax not implemented for this class")

    @property
    def fieldnames(self) :
        return set([elem["field"] for elem in self._fields.values()])

    def write_field(*args,**kwargs) :
        raise BFileError("write_field not implemented for this class")

    def read_field(*args,**kwargs) :
        raise BFileError("read_field not implemented for this class")

    def write_header(self,*args,**kwargs) :
        raise BFileError("write_header not implemented for this class")

    def read_header(self,*args,**kwargs) :
        raise BFileError("read_header not implemented for this class")

    def _open_filea_if_necessary(self,field) :
        if self._filea is None :
            self._jdm,self._idm = field.shape
            self._filea = AFile(self._idm,self._jdm,self._basename+".a",
                    self._action,mask=self._mask,real4=self._real4,endian=self._endian)
        else :
            pass

    def close(self):
        self._filea.close()
        self._fileb.close()

    def check_dimensions(self,field) :
        if self._idm != field.shape[1] or self._jdm != field.shape[0] :
            msg = "dimensions do not match for field. Field dim=%dX%d, file dim=%dX%d"%(field.shape[0].field.shape[1], self._idm, self._jdm)
            raise BFileError(msg)

    @property
    def idm(self):
        return self._idm

    @property
    def jdm(self):
        return self._jdm

    @property
    def fields(self) :
        return self._fields

    @classmethod
    def strip_ab_ending(cls,fname) :
        m=re.match(r"^(.*)(\.[ab]$)", fname)
        if m :
            return m.group(1)
        else :
            return fname

    @classmethod
    def check_minmax(cls,w,mydict) :
#      if (abs(hmina(k)-hminb).gt.abs(hminb)*1.e-4 .or.
#                                &             abs(hmaxa(k)-hmaxb).gt.abs(hmaxb)*1.e-4      ) then
        testmin = numpy.abs(w.min()-mydict["min"]) > numpy.abs(mydict["min"])*1.e-4 
        testmax = numpy.abs(w.max()-mydict["max"]) > numpy.abs(mydict["max"])*1.e-4 
        if testmin or testmax :
            msg="File a b values inconsistent: Field %s at %d, .a min/max: %12.6g %12.6g  .a min/max: %12.6g  %12.6g " %(
                mydict["field"],mydict["k"],w.min(),w.max(),mydict["min"],mydict["max"])
            raise ValueError(msg)

    @property
    def basename(self) :
        return self._basename

class ABFileBathy(ABFile) :
    """ Class for doing input/output on pairs of hycom .a and .b files. This is for bathymetry files"""
    def __init__(self,basename,action,mask=True,real4=True,endian="big",idm=None,jdm=None) :

        super(ABFileBathy,self).__init__(basename,action,mask=mask,real4=real4,endian=endian)
        if action == "r" :
            if idm != None and jdm != None:
                self._idm=idm
                self._jdm=jdm
                self.read_header()
                self.read_field_info()
                self._open_filea_if_necessary(numpy.zeros((self._jdm,self._idm)))
            else :
                raise BFileError("ABFileBathy opened as read, but idm and jdm not provided")
        else :
            self.write_header()

    def write_header(self) :
        self._fileb.write("Bathymetry prepared by python modeltools package\n")
        self._fileb.write("\n")
        self._fileb.write("\n")
        self._fileb.write("\n")
        self._fileb.write("\n")

    def read_header(self) :
        self._header=[]
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())

    def read_field_info(self) :
        fieldkeys=["field","min","max"]
        self._fields={}
        line=self.readline().strip()
        i=0
        while line :
            m = re.match(r"^min,max[ ]+(.*)[ ]*=(.*)",line)
            if m :
                self._fields[i] = {}
                self._fields[i]["field"] = m.group(1).strip()
                elem = [elem.strip() for elem in m.group(2).split() if elem.strip()]
                self._fields[i]["min"] = float(elem[0])
                self._fields[i]["max"] = float(elem[1])
            i+=1
            line=self.readline().strip()

    def write_field(self,field,mask) : 
        self._open_filea_if_necessary(field)
        hmin,hmax = self._filea.writerecord(field,mask,record=None)
        self._fileb.write("min,max %s =%16.5f%16.5f\n"%("depth",hmin,hmax))

    def read_field(self,fieldname) :
        """ Read field corresponding to fieldname and level from bathy file"""
        #print self._fields
        record = None
        for i,d in self._fields.items() :
            if d["field"] == fieldname :
                record=i
        if record  is not None :
            w = self._filea.read_record(record) 
        else :
            raise RuntimeError(f"cannot find field {fieldname} in file {self.basename}")
        return w

    def bminmax(self,fieldname) :
        record=None
        for i,d in self._fields.items() :
            if d["field"] == fieldname :
                record=i
        if record  is not None :
            ret = (self._fields[i]["min"],self._fields[i]["max"])
        else :
            ret = (None,None)
        return ret

class ABFileRmu(ABFile) :
    """ Class for doing input/output on pairs of hycom .a and .b files. This is for nesting/relax rmu files"""
    def __init__(self,basename,action,mask=False,real4=True,endian="big",idm=None,jdm=None,cline1="",cline2="") :

        super(ABFileRmu,self).__init__(basename,action,mask=mask,real4=real4,endian=endian)
        self._cline1=cline1
        self._cline2=cline2
        if action == "w" :
            pass
        else :
            self.read_header()
            self.read_field_info()
            self._open_filea_if_necessary(numpy.zeros((self._jdm,self._idm)))

    def write_header(self) :
        self._fileb.write("%s\n"%self._cline1)
        self._fileb.write("%s\n"%self._cline2)
        self._fileb.write("\n")
        self._fileb.write("\n")
        self._fileb.write("i/jdm =%5d %5d\n"%(self._idm,self._jdm))

    def read_header(self) :
        self._header=[]
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        m = re.match(r"i/jdm[ ]*=[ ]*([0-9]+)[ ]+([0-9]+)",self._header[4].strip())
        if m :
            self._idm = int(m.group(1))
            self._jdm = int(m.group(2))
        else :
            raise  AFileError("Unable to parse idm, jdm from header. File=%s, Parseable string=%s"%(
                    self._filename, self._header[4].strip()))

    def read_field_info(self) :
        fieldkeys=["field","min","max"]
        self._fields={}
        line=self.readline().strip()
        i=0
        while line :
            m = re.match(r"^min,max[ ]+(.*)[ ]*=(.*)",line)
            if m :
                self._fields[i] = {}
                self._fields[i]["field"] = m.group(1).strip()
                elem = [elem.strip() for elem in m.group(2).split() if elem.strip()]
                self._fields[i]["min"] = float(elem[0])
                self._fields[i]["max"] = float(elem[1])
            i+=1
            line=self.readline().strip()

    def write_field(self,field,mask,fieldname,fmt="%16.8g") :
        self._open_filea_if_necessary(field)
        if self._firstwrite :
            self._jdm,self._idm=field.shape
            self._firstwrite=False
            self.write_header()
        self.check_dimensions(field)
        hmin,hmax = self._filea.writerecord(field,mask)
        fmtstr="%%4s:  min,max =%s %s\n"%(fmt,fmt)
        self._fileb.write(fmtstr%(fieldname,hmin,hmax))

    def read_field(self,fieldname) :
        """ Read field corresponding to fieldname and level from bathy file"""
        #print self._fields
        record = None
        for i,d in self._fields.items() :
            if d["field"] == fieldname :
                record=i
        if record  is not None :
            w = self._filea.read_record(record)
        else :
            raise RuntimeError(f"cannot find field {fieldname} in file {self.basename}")
        return w

    def bminmax(self,fieldname) :
        record=None
        for i,d in self._fields.items() :
            if d["field"] == fieldname :
                record=i
        if record  is not None :
            ret = (self._fields[i]["min"],self._fields[i]["max"])
        else :
            ret = (None,None)
        return ret

class ABFileGrid(ABFile) :
    """ Class for doing input/output on pairs of hycom .a and .b files. This is for grid files"""
    fieldkeys=["min","max"]
    def __init__(self,basename,action,mask=False,real4=True,endian="big",mapflg=-1) :

        super(ABFileGrid,self).__init__(basename,action,mask=mask,real4=real4,endian=endian)
        self._mapflg=mapflg

        if action == "w" :
            pass
        else :
            self.read_header()
            self.read_field_info()
            self._open_filea_if_necessary(numpy.zeros((self._jdm,self._idm)))

    def read_header(self) :
        item,self._idm     = self.scanitem(item="idm",conversion=int)
        item,self._jdm     = self.scanitem(item="jdm",conversion=int)
        item,self._mapflg = self.scanitem(item="mapflg",conversion=int)

    def write_header(self) :
        self.writeitem("idm",self._idm)
        self.writeitem("jdm",self._jdm)
        self.writeitem("mapflg",self._mapflg)

    def read_field_info(self) :
        # Get list of fields from .b file
        #plon:  min,max =        -179.99806         179.99998
        #plat:  min,max =         -15.79576          89.98227
        #...
        self._fields={}
        line=self.readline().strip()
        i=0
        while line :
            fieldname = line[0:4]

            self._fields[i]={}
            self._fields[i]["field"]=fieldname
            elems = re.split("[ =]+",line)
            self._fields[i]["min"]=elems[2]
            self._fields[i]["max"]=elems[3]
            for k in self.fieldkeys :
                self._fields[i][k] = float(self._fields[i][k])
            i+=1
            line=self.readline().strip()

    def read_field(self,fieldname) :
        """ Read field corresponding to fieldname and level from archive file"""
        record = None
        for i,d in self._fields.items() :
            if d["field"] == fieldname :
                record=i
        if record  is not None :
            w = self._filea.read_record(record) 
        else :
            raise RuntimeError(f"cannot find field {fieldname} in file {self.basename}")
        return w

    def write_field(self,field,mask,fieldname,fmt="%16.8g") :
        self._open_filea_if_necessary(field)
        if self._firstwrite :
            self._jdm,self._idm=field.shape
            self._firstwrite=False
            self.write_header()
        self.check_dimensions(field)
        hmin,hmax = self._filea.writerecord(field,mask)
        fmtstr="%%4s:  min,max =%s %s\n"%(fmt,fmt)
        self._fileb.write(fmtstr%(fieldname,hmin,hmax))

    def bminmax(self,fieldname) :
        record=None
        for i,d in self._fields.items() :
            if d["field"] == fieldname :
                record=i
        if record  is not None :
            ret = (self._fields[i]["min"],self._fields[i]["max"])
        else :
            ret = (None,None)
        return ret

class ABFileArchv(ABFile) :
    """ Class for doing input/output on pairs of hycom .a and .b files. This is for archv files"""
    fieldkeys=["field","step","day","k","dens","min","max"]
    def __init__(self,basename,action,mask=True,real4=True,endian="big",
            iversn=None,iexpt=None,yrflag=None,cline1="",cline2="",cline3="") :

        self._cline1=cline1
        self._cline2=cline2
        self._cline3=cline3
        self._iversn=iversn
        self._iexpt =iexpt 
        self._yrflag=yrflag

        super(ABFileArchv,self).__init__(basename,action,mask=mask,real4=real4,endian=endian)
        if self._action == "r" :
            self.read_header() # Sets internal metadata. Overrides those on input
            self.read_field_info()
            self._open_filea_if_necessary(numpy.zeros((self._jdm,self._idm)))
        elif self._action == "w" :
            ## Need to test if idm, jdm, etc is set at this stage
            #raise NotImplementedError,"ABFileArchv writing not implemented"
            pass

    def read_header(self) :
        self._header=[]
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())

        item,self._iversn = self.scanitem(item="iversn",conversion=int)
        item,self._iexpt  = self.scanitem(item="iexpt",conversion=int)
        item,self._yrflag = self.scanitem(item="yrflag",conversion=int)
        item,self._idm     = self.scanitem(item="idm",conversion=int)
        item,self._jdm     = self.scanitem(item="jdm",conversion=int)

    def read_field_info(self) :
        # Get list of fields from .b file
        #field         time step  model day  k  dens          min                  max
        #montg1    =        67392     351.000  1 25.000    0.0000000E+00    0.0000000E+00
        #
        self._fields={}
        line=self.readline()
        line=self.readline().strip()
        i=0
        while line :
            elems = re.split("[ =]+",line)
            self._fields[i] = dict(zip(self.fieldkeys,[el.strip() for el in elems]))
            for k in self.fieldkeys :
                if k in ["min","max","dens","day"] :
                    self._fields[i][k] = float(self._fields[i][k])
                elif k in ["k","step"] :
                    self._fields[i][k] = int(self._fields[i][k])
            i+=1
            line=self.readline().strip()

    def read_field(self,fieldname,level) :
        """ Read field corresponding to fieldname and level from archive file"""
        record = None
        for i,d in self._fields.items() :
            if d["field"] == fieldname and level == d["k"] :
                record=i
        if record  is not None :
            r = self._filea.read_record(record)
            w = r.data
            w[r.mask] = numpy.nan
        else :
            raise RuntimeError(f"cannot find field {fieldname} in file {self.basename}")
        return w

    def write_header(self) :
        self._fileb.write(self._cline1+"\n")
        self._fileb.write(self._cline2+"\n")
        self._fileb.write(self._cline3+"\n")
        self._fileb.write("12345678901234567890123456789012345678901234567890123456789012345678901234567890\n")
        self.writeitem("iversn",self._iversn)
        self.writeitem("iexpt" ,self._iexpt)
        self.writeitem("yrflag",self._yrflag)
        self.writeitem("idm"    ,self._idm)
        self.writeitem("jdm"    ,self._jdm)
        self._fileb.write("field         time step  model day  k  dens          min                  max\n")

    def write_field(self,field,mask,fieldname,time_step,model_day,k,dens) :
        self._open_filea_if_necessary(field)
        if self._firstwrite :
            self._jdm,self._idm=field.shape
            self._firstwrite=False
            self.write_header()
        self.check_dimensions(field)
        hmin,hmax = self._filea.writerecord(field,mask)
        fmtstr="%-9s=%11d%11.2f%3d%7.3f%16.7E%16.7E\n"
        self._fileb.write(fmtstr%(fieldname,time_step,model_day,k,dens,hmin,hmax))

# Mostafa
    def get_fields(self):
        return self._fields

    @property
    def fieldlevels(self) :
        return set([elem["k"] for elem in self._fields.values()])

    def bminmax(self,fieldname,k) :
        record=None
        for i,d in self._fields.items() :
            if d["field"] == fieldname and d["k"] == k:
                record=i
        if record  is not None :
            ret = (self._fields[i]["min"],self._fields[i]["max"])
        else :
            ret = (None,None)
        return ret

    @property
    def iversn(self) : return self._iversn

    @property
    def iexpt(self) : return self._iexpt

    @property
    def yrflag(self) : return self._yrflag

class ABFileForcing(ABFile) :
    """ Class for doing input/output on pairs of hycom .a and .b files. This is for forcing files"""
    fieldkeys=["field","min","max"]
    def __init__(self,basename,action,mask=False,real4=True,endian="big", idm=None,jdm=None,
                     cline1="",cline2=""):

        super(ABFileForcing,self).__init__(basename,action,mask=mask,real4=real4,endian=endian)
        self._cline1=cline1
        self._cline2=cline2
        if self._action[0] == "r" :
            self.read_header()
            self.read_field_info()
            self._open_filea_if_necessary(numpy.zeros((self._jdm,self._idm)))
        elif self._action == "w":
            self._idm = idm
            self._jdm = jdm

    def find_record(self, fieldname, dtime1):
        record = None
        for i,d in self._fields.items() :
            if d["field"] == fieldname and d["dtime1"] == dtime1:
                record=i
        return record

    def write_bfile_info(self) :
        self._fileb.seek(0)
        self.write_header()
        for i in range(len(self._fields)):
            fieldname = self._fields[i]["field"]
            dtime1 = self._fields[i]["dtime1"]
            rdtime = self._fields[i]["range"]
            hmin = self._fields[i]["min"]
            hmax = self._fields[i]["max"]
            self._fileb.write("%s:dtime1,range = %12.4f%12.4f,%14.6e%14.6e\n"%(fieldname,dtime1,rdtime,hmin,hmax))

    def read_header(self) :
        self._header=[]
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._cline1=self._header[0].strip()
        self._cline2=self._header[1].strip()
        m = re.match("i/jdm[ ]*=[ ]*([0-9]+)[ ]+([0-9]+)",self._header[4].strip())
        if m :
            self._idm = int(m.group(1))
            self._jdm = int(m.group(2))
        else :
            raise  AFileError("Unable to parse idm, jdm from header. File=%s, Parseable string=%s"%(
                    self._filename, self._header[4].strip()))

    def write_header(self) :
        self._fileb.write(self._cline1.strip()+"\n")
        self._fileb.write(self._cline2.strip()+"\n")
        self._fileb.write("\n")
        self._fileb.write("\n")
        self._fileb.write("i/jdm =%5d %5d\n"%(self._idm,self._jdm))

    def write_field(self,field,mask,fieldname,dtime1,rdtime) :
        self._open_filea_if_necessary(field)
        if self._firstwrite :
            self._jdm,self._idm=field.shape
            self._firstwrite=False
            self.write_header()
        self.check_dimensions(field)
        hmin,hmax = self._filea.writerecord(field,mask)
        self._fileb.write("%s:dtime1,range = %12.4f%12.4f,%14.6e%14.6e\n"%(fieldname,dtime1,rdtime,hmin,hmax))

    def overwrite_field(self, field, mask, fieldname, dtime1) :
        assert self._action == "r+"
        record = self.find_record(fieldname, dtime1)
        assert record is not None, f"Do not find dtime1={dtime1} in current file, existing"
        self._open_filea_if_necessary(field)
        self.check_dimensions(field)
        hmin,hmax = self._filea.writerecord(field, mask, record=record)
        self._fields[record]["min"] = hmin
        self._fields[record]["max"] = hmax
        self.write_bfile_info()

    def read_field_info(self) :
        # Get list of fields from .b file
        #plon:  min,max =        -179.99806         179.99998
        #plat:  min,max =         -15.79576          89.98227
        #...
        self._fields={}
        line=self.readline().strip()
        i=0
        while line :
            m = re.match(r"^(.*):dtime1,range[ ]*=[ ]+([0-9\-\.e+]+)[ ]+([0-9\-\.e+]+)[ ]*,[ ]*([0-9\-\.e+]+)[ ]*([0-9\-\.e+]+)",line)
            if m :
                self._fields[i] = {}
                self._fields[i]["field"]  = m.group(1).strip()
                self._fields[i]["dtime1"] = float(m.group(2).strip())
                self._fields[i]["range"]  = float(m.group(3).strip())
                self._fields[i]["min"]  = float(m.group(4).strip())
                self._fields[i]["max"]  = float(m.group(5).strip())
            else :
                raise NameError("cant parse forcing field in "+self._basename)
            i+=1
            line=self.readline().strip()

    def read_field(self,field,dtime1) :
        """ Read field corresponding to fieldname and level from archive file"""
        elems = [ (k,v["dtime1"]) for k,v in self._fields.items() if v["field"] == field]
        dist = numpy.array([elem[1]-dtime1 for elem in elems])
        ##tolerance for max dist is 1 day
        if numpy.min(numpy.abs(dist)) > 1.0 :
            raise RuntimeError(f"cannot find matching time {dtime1} in file {self.basename}, closest time dist is {dist}.")
        i =numpy.argmin(numpy.abs(dist))
        rec,dt = elems[i]
        w = self._filea.read_record(i) 
        #print w
        return w#,dt

    def bminmax(self,fieldname,dtime1) :
        record=None
        for i,d in self._fields.items() :
            if d["field"] == fieldname and d["dtime1"] == dtime1:
                record=i
        if record  is not None :
            ret = (self._fields[i]["min"],self._fields[i]["max"])
        else :
            ret = (None,None)
        return ret

    @property
    def field_times(self) :
        return set([elem["dtime1"] for elem in self._fields.values()])

class ABFileRiver(ABFile) :
    """ Class for doing input/output on pairs of hycom .a and .b files. This is for river forcing files"""
    """ At the moment, only write river function is implemented. You can read it by using 'loada' subroutines """
    fieldkeys=["field","min","max"]
    def __init__(self,basename,action,mask=False,real4=True,endian="big", idm=None,jdm=None,
                     cline1="",cline2=""):

        super(ABFileRiver,self).__init__(basename,action,mask=mask,real4=real4,endian=endian)
        self._idm=idm
        self._jdm=jdm
        self._cline1=cline1
        self._cline2=cline2
        if action == "w" :
            pass
        else :
            quit()

    def write_header(self) :
        self._fileb.write(self._cline1.strip()+"\n")
        self._fileb.write(self._cline2.strip()+"\n")
        self._fileb.write("\n")
        self._fileb.write("\n")
        self._fileb.write("i/jdm =%5d %5d\n"%(self._idm,self._jdm))

    def write_field(self,field,mask,fieldname,month) :
        self._open_filea_if_necessary(field)
        self.check_dimensions(field)
        hmin,hmax = self._filea.writerecord(field,mask)
        self._fileb.write(" %s: month,range = %s%14.6E%14.6E\n"%(fieldname,str(month).zfill(2),hmin,hmax))

class ABFileRestart(ABFile) :
    """ Class for doing input/output on pairs of hycom .a and .b files. This is for restart files"""
    fieldkeys=["field","step","day","k","dens","min","max"]
    def __init__(self,basename,action,mask=False,real4=True,endian="big",
            iversn=None,iexpt=None,yrflag=None,idm=None,jdm=None) :

        super(ABFileRestart,self).__init__(basename,action,mask=mask,real4=real4,endian=endian)
        if self._action[0] == "r" :
            if idm != None and jdm != None:
                self._idm=idm
                self._jdm=jdm
                self.read_header() # Sets internal metadata. Overrides those on input
                self.read_field_info()
                self._open_filea_if_necessary(numpy.zeros((self._jdm,self._idm)))
            else :
                raise BFileError("ABFile opened as read, but idm and jdm not provided")
        elif self._action == "w" :
##            # Need to test if idm, jdm, etc is set at this stage
##            raise NotImplementedError,"ABFileRestart writing not implemented
            pass

    def read_header(self) :
        #RESTART2: iexpt,iversn,yrflag,sigver =     990     22      3      2
        #RESTART2: nstep,dtime,thbase =        19636860     40910.12500000000            34.00000000000000
        self._header=[]
        self._header.append(self.readline())
        self._header.append(self.readline())

        m=re.match(r"RESTART2: iexpt,iversn,yrflag,sigver[ ]*=[ ]*([0-9]+)[ ]+([0-9]+)[ ]+([0-9]+)[ ]+([0-9]+)",self._header[0])
        self._iexpt=int(m.group(1))
        self._iversn=int(m.group(2))
        self._yrflag=int(m.group(3))
        self._sigver=int(m.group(4))
        m2=re.match(r"RESTART2: nstep,dtime,thbase[ ]*=[ ]*([^ ]+)[ ]+([^ ]+)[ ]+([^ ]+)",self._header[1])
        self._nstep=int(m2.group(1))
        self._dtime=float(m2.group(2))
        self._thbase=float(m2.group(3))

    def read_field_info(self) :
        #u         : layer,tlevel,range =    1  1     -9.2374402E-01    6.0595530E-01
        self._fields={}
        line=self.readline().strip()
        i=0
        while line :
            m = re.match(r"^([a-zA-Z0-9_ ]+): layer,tlevel,range =[ ]*([^ ]+)[ ]+([^ ]+)[ ]+([^ ]+)[ ]+([^ ]+)",line)
            if m :
                self._fields[i] = {}
                self._fields[i]["field"] = m.group(1).strip()
                self._fields[i]["k"] = int(m.group(2))
                self._fields[i]["tlevel"] = int(m.group(3))
                self._fields[i]["min"] = float(m.group(4))
                self._fields[i]["max"] = float(m.group(5))
            else :
                raise NameError("unable to parse line %s"%line)
            #print i,line
            i+=1
            line=self.readline().strip()

    def write_bfile_info(self) :
        self._fileb.seek(0)
        ##header
        self._fileb.write("RESTART2: iexpt,iversn,yrflag,sigver = %6d%6d%6d%6d\n"%(self._iexpt,self._iversn,self._yrflag,self._sigver))
        self._fileb.write("RESTART2: nstep,dtime,thbase = %12d%19.10f%24.13f\n"%(self._nstep,self._dtime,self._thbase))
        ##field info
        for i in range(len(self._fields)):
            fieldname = self._fields[i]["field"]
            k = self._fields[i]["k"]
            time = self._fields[i]["tlevel"]
            hmin = self._fields[i]["min"]
            hmax = self._fields[i]["max"]
            fmtstr="%-8s: layer,tlevel,range = %3d%3d%18.7E%16.7E\n"
            self._fileb.write(fmtstr%(fieldname,k,time,hmin,hmax))

    def find_record(self, fieldname, level, tlevel):
        record = None
        for i,d in self._fields.items() :
            if d["field"] == fieldname and level == d["k"] and d["tlevel"] == tlevel :
                record=i
        return record

    def read_field(self,fieldname,level,tlevel=1,mask=None) :
        """ Read field corresponding to fieldname and level from archive file"""
        record = self.find_record(fieldname, level, tlevel)
        if record  is not None :
            w = self._filea.read_record(record)
            ABFile.check_minmax(w,self._fields[record]) # Always do this check
            fld = w.data
            if mask is None:
                fld[w.mask] = numpy.nan
            else:
                fld[mask] = numpy.nan
        else :
            raise RuntimeError(f"cannot find field {fieldname} in file {self.basename}")
        return fld

    def overwrite_field(self, field, mask, fieldname, level, tlevel=1) :
        assert self._action == "r+"
        record = self.find_record(fieldname, level, tlevel)
        assert record is not None, "Do not find record in current file, existing"
        self._open_filea_if_necessary(field)
        self.check_dimensions(field)
        hmin,hmax = self._filea.writerecord(field, mask, record=record)
        self._fields[record]["min"] = hmin
        self._fields[record]["max"] = hmax
        self.write_bfile_info()

# THE RESTART WRITING - a later addition to KAL's original abfile.py
#  caglar: other classes in abfile.py code call 'write_header' within 'write_field'.
#             Because of many inputs in the header, I decided to do it separately.
#             Make sure that you call 'write_header' only once before calling the rest of 'write_field'
#  example lines:
#  newfile = 'merged'
#  new_abfile = abfile.ABFileRestart(newfile,"w",idm=I2,jdm=I1)
#  new_abfile.write_header(iexpt,iversn,yrflag,sigver,nstep,dtime,thbase)
#  ~~ for loop for variables
#      new_abfile.write_field(field,None,fieldname,k,t)  # None stands for mask
#  new_abfile.close()
##YY: the above is suitable for writing a new file from scratch (copy and paste)
##overwrite_field will be good for rewriting only a field in existing files

    def write_header(self,iexpt,iversn,yrflag,sigver,nstep,dtime,thbase) :
        self._fileb.write("RESTART2: iexpt,iversn,yrflag,sigver = %6d%6d%6d%6d\n"%(iexpt,iversn,yrflag,sigver))
        self._fileb.write("RESTART2: nstep,dtime,thbase = %12d%19.10f%24.13f\n"%(nstep,dtime,thbase))

    def write_field(self,field,mask,fieldname,k,time,record=None) :
        self._open_filea_if_necessary(field)
        self.check_dimensions(field)
        hmin,hmax = self._filea.writerecord(field,mask,record)
        fmtstr="%-8s: layer,tlevel,range = %3d%3d%18.7E%16.7E\n"
        self._fileb.write(fmtstr%(fieldname,k,time,hmin,hmax))

    @property
    def fieldlevels(self) :
        return set([elem["k"] for elem in self._fields.values()])

    def bminmax(self,fieldname,k) :
        record=None
        for i,d in self._fields.items() :
            if d["field"] == fieldname and d["k"] == k:
                record=i
        if record  is not None :
            ret = (self._fields[i]["min"],self._fields[i]["max"])
        else :
            ret = (None,None)
        return ret

class ABFileRelax(ABFile) :
    """ Class for doing input/output on pairs of hycom .a and .b files. This is for hybrid coord relaxation data used by hycom """
    fieldkeys=["field","layer","dens","min","max"]
    def __init__(self,basename,action,mask=False,real4=True,endian="big", idm=None,jdm=None,
                     cline1="",cline2=""):
        super(ABFileRelax,self).__init__(basename,action,mask=mask,real4=real4,endian=endian)
        self._cline1=cline1
        self._cline2=cline2
        if action == "w" :
            raise NotImplementedError
            pass
        else :
            self.read_header()
            self.read_field_info()
            self._open_filea_if_necessary(numpy.zeros((self._jdm,self._idm)))

#woa2013 Climatology
#Expt 99.0  nhybrd=32 nsigma=14 ds00= 0.50 dp00= 3.00 dp00x= 450.0 dp00f=1.180
#Layered averages w.r.t. Sigma-2,  levtop=1 (sigver= 2)
#Potential Temperature
#i/jdm =  800  760
    def read_header(self) :
        self._header=[]
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        #print "test",self._header
        m = re.match(r"i/jdm[ ]*=[ ]*([0-9]+)[ ]+([0-9]+)",self._header[4].strip())
        if m :
            self._idm = int(m.group(1))
            self._jdm = int(m.group(2))
        else :
            raise  BFileError("Unable to parse idm, jdm from header. File=%s, Parseable string=%s"%(
                    self._basename, self._header[4].strip()))
        #print self._idm,self._jdm

    def read_field_info(self) :
        # Typical line
        #tem: month,layer,dens,range = 01  01 28.100  -3.3520899E+00    1.0784083E+01
        self._fields={}
        line=self.readline().strip()
        i=0
        while line :
            m = re.match(r"^([a-z_ ]+):[ ]*month[ ]*,[ ]*layer[ ]*,[ ]*dens[ ]*,[ ]*range[ ]*=[ ]*([^ ]+)[ ]+([^ ]+)[ ]+([^ ]+)[ ]+([^ ]+)[ ]+([^ ]+)",line)
            if m :
                self._fields[i] = {}
                self._fields[i]["field"] = m.group(1).strip()
                self._fields[i]["month"] = int(m.group(2))
                self._fields[i]["k"] = int(m.group(3))
                self._fields[i]["dens"] = float(m.group(4))
                self._fields[i]["min"] = float(m.group(5))
                self._fields[i]["max"] = float(m.group(6))
                #print self._fields[i]
            else :
                raise NameError("unable to parse line %s"%line)
            #print i,line
            i+=1
            line=self.readline().strip()

    def read_field(self,fieldname,level,month) :
        """ Read field corresponding to fieldname and level from archive file"""
        record = None
        for i,d in self._fields.items() :
            if d["field"] == fieldname and level == d["k"] and d["month"] == month :
                record=i
        if record  is not None :
            w = self._filea.read_record(record) 
            ABFile.check_minmax(w,self._fields[record]) # Always do this check
        else :
            raise RuntimeError(f"cannot find field {fieldname} at level {level} month {month} in file {self.basename}")
        return w

class ABFileRelaxZ(ABFile) :
    """ Class for doing input/output on pairs of hycom .a and .b files. This is for z level data used by hycom relax routine """
    fieldkeys=["field","depth","min","max"]
    def __init__(self,basename,action,mask=False,real4=True,endian="big", idm=None,jdm=None,
                     cline1="",cline2=""):
        super(ABFileRelaxZ,self).__init__(basename,action,mask=mask,real4=real4,endian=endian)
        self._cline1=cline1
        self._cline2=cline2
        if action == "w" :
            pass
        else :
            self.read_header()
            self.read_field_info()
            self._open_filea_if_necessary(numpy.zeros((self._jdm,self._idm)))

    def read_header(self) :
        self._header=[]
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._header.append(self.readline())
        self._cline1=self._header[0].strip()
        self._cline2=self._header[1].strip()
        m = re.match(r"i/jdm[ ]*=[ ]*([0-9]+)[ ]+([0-9]+)",self._header[4].strip())
        if m :
            self._idm = int(m.group(1))
            self._jdm = int(m.group(2))
        else :
            raise  AFileError("Unable to parse idm, jdm from header. File=%s, Parseable string=%s"%(
                    self._filename, self._header[4].strip()))

    def write_header(self) :
        self._fileb.write(self._cline1.strip()+"\n")
        self._fileb.write(self._cline2.strip()+"\n")
        self._fileb.write("\n")
        self._fileb.write("\n")
        self._fileb.write("i/jdm =%5d %5d\n"%(self._idm,self._jdm))

    def write_field(self,field,mask,fieldname,depth) :
        self._open_filea_if_necessary(field)
        if self._firstwrite :
            self._jdm,self._idm=field.shape
            self._firstwrite=False
            self.write_header()
        self.check_dimensions(field)
        hmin,hmax = self._filea.writerecord(field,mask)
        self._fileb.write("%s: depth,range = %12.4f %14.6e%14.6e\n"%(fieldname,depth,hmin,hmax))

    def read_field_info(self) :
        # Get list of fields from .b file
        #plon:  min,max =        -179.99806         179.99998
        #plat:  min,max =         -15.79576          89.98227
        #...
        self._fields={}
        line=self.readline().strip()
        i=0
        while line :
            m = re.match(r"^(.*):[ ]*depth,[ ]*range[ ]*=[ ]*([0-9\-\.e+]+)[ ]+([0-9\-\.e+]+)[ ]+([0-9\-\.e+]+)",line)
            if m :
                self._fields[i] = {}
                self._fields[i]["field"]  = m.group(1).strip()
                self._fields[i]["depth"] = float(m.group(2).strip())
                self._fields[i]["min"]  = float(m.group(3).strip())
                self._fields[i]["max"]  = float(m.group(4).strip())
            else :
                raise NameError("cant parse relax field in "+self._basename)
            i+=1
            line=self.readline().strip()

    def read_field(self,field,dtime1) :
        """ Read field corresponding to fieldname and level from archive file"""
        elems = [ (k,v["dtime1"]) for k,v in self._fields.items() if v["field"] == field]
        dist = numpy.array([elem[1]-dtime1 for elem in elems])
        ##tolerance for max dist is 1 day
        if numpy.min(numpy.abs(dist)) > 1.0 :
            raise RuntimeError(f"cannot find matching time {dtime1} in file {self.basename}, closest time dist is {dist}.")
        i =numpy.argmin(numpy.abs(dist))
        rec,dt = elems[i]
        w = self._filea.read_record(i) 
        #print w
        return w#,dt

    def bminmax(self,fieldname,depth) :
        record=None
        for i,d in self._fields.items() :
            if d["field"] == fieldname and d["depth"] == depth:
                record=i
        if record  is not None :
            ret = (self._fields[i]["min"],self._fields[i]["max"])
        else :
            ret = (None,None)
        return ret

def write_bathymetry(exp,version,d,threshold) :
    myfile="depth_%s_%02d"%(exp,version)
    #regf = ABFileBathy(myfile,"w",idm=d.shape[0],jdm=d.shape[1],mask=True)
    #print type(d),d.min(),d.max()
    #mask=d <= threshold
    #regf.write_field(d,mask)
    #No mask!
    tmp = numpy.copy(d)
    mask = tmp <= threshold
    tmp[mask] = AFile.huge
    regf = ABFileBathy(myfile,"w",idm=d.shape[0],jdm=d.shape[1],mask=True)
    regf.write_field(tmp,mask)
    regf.close()

def write_regional_grid(datadict,endian="big") :
    regf = ABFileGrid("regional.grid","w",mapflg=-1,endian=endian)
    for key in grid_ordered_fieldnames :
        regf.write_field(datadict[key],datadict[key],key)
    regf.close()

def read_regional_grid(endian="big") :
    regf = ABFileGrid("regional.grid","r",endian=endian)
    res={}
    for fldname in grid_ordered_fieldnames :
        res[fldname] = regf.read_field(fldname)
    regf.close()
    return res

def write_diag_nc(datadict,fname="hycom_grid.nc") :
    try :
        import netCDF4
    except :
        print("Unable to import netCDF4 module - will not write diag to %s"%fname)
        return
    tmp = datadict["plon"]
    ## Create netcdf file with all  stages for analysis
    ncid = netCDF4.Dataset(fname,"w")
    ncid.createDimension("idm",tmp.shape[1])
    ncid.createDimension("jdm",tmp.shape[0])
    for i in datadict.keys() :
        ncid.createVariable(i,"f8",("jdm","idm"))
    for i in datadict.keys() :
        ncid.variables[i][:]=datadict[i][:]
    ncid.close()




"""
This module contain classes to handle grid with various geometries and spatial structures.

Classes
-------
- :class:`Grid`: A factory class for 2D grids.
- :class:`GridBase`: Base class for handling 2D grids.
- :class:`RegularGrid`: Class for handling 2D regular grid.
- :class:`IrregularGrid`: Class for handling 2D irregular grid (triangular mesh).
- :class:`Grid1D`: Class for handling 1D grid.

Map projections
---------------
The projection is typically handled by a :class:`pyproj.Proj` object, see `list of supported projections <https://proj.org/en/9.2/operations/projections/index.html>`_.
However, if you are working with some unsupported grid geometries, you can also define a custom `proj` function.
What :class:`Grid` needs from :func:`proj` is the conversion between longitude,latitude and the x,y coordinates, and the inverse, i.e.

>>> x, y = proj(lon, lat)
>>> lon, lat = proj(x, y, inverse=True)

Coordinates
-----------
The coordinates `x,y` has the same dimensions as the 2D grid.

For a regular grid, when we have a scalar field indexed by `fld[j,i]`, its coordinate is `x[j,i], y[j,i]`. Note that for Python the rightmost dimension is the quickest (row index first), if you have data that is column index first `fld[i,j]` then you will need to transpose it before using `Grid`.

So far we only support regular grid x,y where all the `x[j,:]` and all the `y[:,i]` are identical, namely `x, y = np.meshgrid(xc, yc)` where `xc,yc` are the 1D coordinates arrays. If you have a rotated x,y plane, you need to let the projection take care of the rotation first.

Cyclic boundary condition is supported, by setting the `cyclic_dim` parameter you can tell `Grid` which dimension(s) should wrap around when the coordinates fall outside of the given range (done by `wrap_cyclic_xy()`)

The longitude,latitude (geodetic) grid is special since it is the only projection not on a flat surface. For this `+proj=longlat` projection we support longitude conventions both 0\~360 and -180\~180 as long as its unit is "degrees east", you shall set `cyclic_dim='x'` if x is your longitude dimension. The latitude has units "degrees north". We also support special treatment of the two poles (lat=-90, 90) at which rotation angle is not well defined. You shall specify which dimension contains the pole and which index in that dimension is the pole. For example, if your data is [lat=40\~90, long=0\~360], you can set `pole_dim='y'` and `pole_index=(-1)`.

For an unstructured mesh, the `x,y` coordinates are for the nodes (vertices of the triangles), the 2D field can be either defined on nodes or elements (triangle faces). If the later is true, `x_elem,y_elem` will be used in handling the field. Note that `x_elem` has different size with `x`.




Examples:
    Define a proj function, e.g. stereographic:

    >>> import pyproj
    >>> proj = pyproj.Proj('+proj=stere +lat_0=90 +lon_0=0')

    Create a regular mesh x, y coordinates, with 10 km resolution, in meter units:

    >>> import numpy as np
    >>> x, y = np.meshgrid(np.arange(-3e6, 3e6, 1e4), np.arange(-3e6, 3e6, 1e4))

    Create a grid instance:

    >>> from NEDAS.grid import Grid
    >>> grid = Grid(proj, x, y)

`self.nx`, `self.ny` are the grid dimension for regular grids.

`self.xmin`, `self.xmax`, `self.ymin`, `self.ymax` define the grid extent (bounding box) for regular grids.

`self.Lx`, `self.Ly` are the grid length in x and y directions.

`self.dx` is the grid spacing (resolution), for unstructured mesh it is the averaged edge length of the triangles. `self.dy` is also provided for regular grid, although it is typical that we use `dy`=`dx`.

`self.mfx`, `self.mfy` are the map factors in x and y directions. Due to map projection the actual length is distorted on the 2D grid plane (see [Tissot's indicatrix](https://en.wikipedia.org/wiki/Tissot%27s_indicatrix) for an illustration). These map factors are useful in computing actual distances on Earth, using `self.dx / self.mfx` and `self.dy / self.mfy`

`self.tri` for unstructured mesh this is the triangle indices, computed by `matplotlib.tri.Triangulation`

`self.x_elem`, `self.y_elem` are the center coordinates for each triangular element.

Typically we use the `Grid` class for conversion of a 2D field from one grid to another, which involves interpolation, rotation of vectors and coarse-graining if the two grid has different resolution. Additionally, we also provide basic tools for visualizing a field on the map. See `tutorials/grid_convert` for some illustrated examples.

Once you created a `grid1` object on which a 2D field `fld1` is defined, and a `grid2` object for the destination grid. You convert `fld1` by first setting

`grid1.set_destination_grid(grid2)`, then

`fld2 = grid1.convert(fld1)`

It is recommended that you initialize a `grid` object, set the destination grid by `set_destination_grid`, then apply the same `grid.convert` to many fields needing conversion. Since the interpolation and rotation matrices are computed by `set_destination_grid` only once, the following application of `grid.convert` can be very efficient.


### 2.1 Interpolation

Two methods are supported for now. `method=linear` does bi-linear interpolation for the regular grids and barycentric interpolation for the triangular meshes. `method=nearest` performs nearest neighbor interpolation.

If you just want to do interpolation, you can also call `grid.interp(fld, x, y, method)` directly, where `fld` is the original field, `x,y` are the target coordinates (can be just one number, or an array, multi-dimension arrays will be flattened), `method` is the interpolation method.

If you only want the indices of given `x,y` coordinates, you can also use `grid.find_index(x, y)`. But note that we return more than just the indices, also some other information. Five things will be returned by `find_index`:

* `inside`: True if the point x,y is inside the grid

* `indices`: `None` for regular grid; triangle index which the `x,y` position falls in.

* `vertices`: the indices of rectangle(4) or triangle(3) vertices the `x,y` position falls in.

* `in_coords`: the internal coordinates of the `x,y` point within the rectangle (`in_x`, `in_y`) or triangle (`in1`, `in2`, `in3`)

* `nearest`: index in `x.flatten()`, `y.flatten()` for the nearest grid point to the given `x,y` position.


### 2.2 Rotation of Vectors

We support both scalar and vector fields. For a vector field, its u- and v- components are stacked in the first dimension, so that `vec_fld` has dimension [2, `ny`, `nx`]. Parameter `is_vector` = `True` will let `grid.convert` know it's operating on a vector field and need to rotate vectors if projection changes.

The rotation matrix is computed by displacing a small amount in x,y directions on each grid point, and project both the original point and displaced point to the new grid, then measuring the displacement and figure out the rotation.

If you just want to rotate vectors, you can apply `grid.rotate_vectors(vec_fld)`.


### 2.3 Coarse-graining

Setting `coarse_grain=True` in `grid.convert` will perform an additional coarse graining (by `grid.coarsen`) during interpolation. When the destination grid has lower resolution than the source grid, more points will fall in the same destination grid element. The coarse-graining will average them to represent the field on the low-resolution grid, instead of interpolating only the nearest source grid points. For a field with small-scale variability on the source grid, the interpolation to low-resolution grid will be sensitive to those sub-grid-scale variability, causing the interpolated value to have representation errors. Taking the average in the low-resolution grid will remedy this issue.

However, if you are dealing with a sparse observing network where you want to interpolate in the source grid to find the observed values, you don't need coarse graining, because observation should pick up sub-grid scale variability in its measurement anyway. Remember to set `coarse_grain=False` if this is the case.


### 2.4 Visualization

To avoid intricate dependencies from installing the `cartopy` package, we instead provide some internal solutions for basic plotting.

`ax` is a `matplotlib.pyplot.Axes` object for plotting.

Scalar fields can be plotted using `grid.plot_field(ax, fld, vmin, vmax, cmap)`, we use `pyplot.pcolor` for regular grids and `pyplot.tripcolor` for unstructured meshes. `vmin`, `vmax` are minimum and maximum values shown for the field. `cmap` is the color map.

Vector fields can be plotted using `grid.plot_vectors(ax, vec_fld, V, L, spacing, num_steps, linecolor, linewidth, showred, ref_xy, refcolor, showhead, headwidth, headlength)`. This replaces the `pyplot.quiver` and provides more control over how a vector plot looks. `vec_fld` has a dimension [2, ...] for the two components. `V` is the velocity scale that is the typical velocity value in this field (same units as `vec_fld`), `L` is the length scale that is the length (in `x,y` units) for vectors with velocity `V`. If you don't provide `V` and `L`, a standard value will be used based on the data provided, which is often good enough. `spacing` controls the density of vectors in the plot, they are spaced with interval `spacing * L`. `num_steps`=1 displays straight vectors (as in `quiver`); >1 lets you display curved trajectories, at each sub-step the velocity is re-interpolated at the new position along the trajectories. `linecolor, linewith` controls the style of vector lines. `show_ref=True` displays the reference vector at position `ref_xy=(x,y)` and with background color `refcolor`. `showhead=True` displays the vector heads with `headwidth, headlength` controlling their styles.


To show the map overlay, you can use `grid.plot_land(ax, color, linecolor, linewidth, showriver, rivercolor, showgrid, dlon, dlat)`. There are several shape files downloaded from [www.naturalearthdata.com](https://www.naturalearthdata.com) are stored in `grid` directory, where the `plot_land` reads and gather `land_data`, `river_data` and `lake_data`. They data are cached after first time use, so making several plots in a role will be more efficient. `color` controls the face color of the landmass, `linecolor` and `linewidth` controls the style of the coastlines. `showriver=True` will plot the rivers and lakes with `rivercolor`. `showgrid=True` turns on the latitude longitude grid at `dlon, dlat` intervals (degrees). The x,y axes will show the coordinate tick marks in its units.

"""

from typing import Optional, Any
import numpy as np
from NEDAS.grid.grid_regular import RegularGrid
from NEDAS.grid.grid_irregular import IrregularGrid

class Grid:
    """
    Factory class for creating a 2D grid object.

    This class serves as a factory for creating either a RegularGrid or an IrregularGrid instance
    based provided parameters. It allows for easy instantiation of different grid types
    without needing to directly interact with the specific grid classes.

    Examples:
        The factory class provides two classmethods to help creating a grid instance more easily.
        To create a regular grid with coordinates x from 0 to 100 and y from 0 to 100, resolution dx=1:

        >>> from NEDAS.grid import Grid
        >>> grid = Grid.regular_grid(None, 0, 100, 0, 100, 1)

        To create a irregular grid with same range of coordinates, in total 2000 grid points randomly placed:

        >>> grid = Grid.random_grid(None, 0, 100, 0, 100, 2000)
    """
    def __new__(cls, proj, x, y, regular=True, bounds=None, cyclic_dim=None, pole_dim=None, pole_index=None,
                distance_type='cartesian', triangles=None, neighbors=None, dst_grid=None):
        if regular:
            return RegularGrid(proj, x, y, bounds, cyclic_dim, distance_type,
                               pole_dim, pole_index, neighbors, dst_grid)
        else:
            return IrregularGrid(proj, x, y, bounds, cyclic_dim, distance_type,
                                 triangles, dst_grid)

    @classmethod
    def regular_grid(cls, proj: Any,
                     xstart: float, xend: float, ystart: float, yend: float,
                     dx: float, centered: bool=False, **kwargs) -> RegularGrid:
        """
        Create a regular grid within specified boundaries.

        Parameters:
            proj (Any): Projection from lon,lat to x,y.
            xstart (float): Lower bound for X coordinates.
            xend (float): Upper bound for X coordinates.
            ystart (float): Lower bound for Y coordinates.
            yend (float): Upper bound for Y coordinates.
            dx (float): Grid spacing.
            centered (bool): Optional, toggle for grid points to be on vertices (False) or in the middle of each grid box (True).  Default is False.
            **kwargs: Additional keyword arguments.

        Returns:
            RegularGrid: A Grid object representing the regular grid.
        """
        dx = float(dx)
        xcoord = np.arange(xstart, xend, dx)
        ycoord = np.arange(ystart, yend, dx)
        x, y = np.meshgrid(xcoord, ycoord)
        if centered:
            x += 0.5*dx  ##move coords to center of grid box
            y += 0.5*dx
        return RegularGrid(proj, x, y, **kwargs)

    @classmethod
    def random_grid(cls, proj: Any,
                    xstart: float, xend: float, ystart: float, yend: float,
                    npoints: int, min_dist: Optional[float]=None, **kwargs) -> IrregularGrid:
        """
        Create a grid with randomly positioned points within specified boundaries.

        Parameters:
            proj (pyproj.Proj): Projection from lon,lat to x,y.
            xstart (float): Lower bound for X coordinates.
            xend (float): Upper bound for X coordinates.
            ystart (float): Lower bound for Y coordinates.
            yend (float): Upper bound for Y coordinates.
            npoints (int): Number of grid points.
            min_dist (float): Optional, minimal distance allowed between each pair of grid points.
            **kwargs: Additional keyword arguments.

        Returns:
            IrregularGrid: A Grid object representing the randomly positioned grid.
        """
        points = []
        ntry = 0
        while len(points) < npoints:
            xp = np.random.uniform(0, 1) * (xend - xstart) + xstart
            yp = np.random.uniform(0, 1) * (yend - ystart) + ystart
            if min_dist is not None:
                near = [p for p in points if abs(p[0]-xp) <= min_dist and abs(p[1]-yp) <= min_dist]
                ntry += 1
                if not near:
                    points.append((xp, yp))
                    ntry = 0
                if ntry > 10:
                    raise RuntimeError(f"tried 10 times, unable to insert more grid points so that dist<{min_dist}")
            else:
                points.append((xp, yp))
        x = np.array([p[0] for p in points])
        y = np.array([p[1] for p in points])
        bounds = [xstart, xend, ystart, yend]
        return IrregularGrid(proj, x, y, bounds=bounds, **kwargs)



import importlib
from typing import Optional
from NEDAS.job_submitters.base import JobSubmitter

registry_by_host = {
    'macos': 'MacOSJobSubmitter',
    'betzy': 'BetzyJobSubmitter',
}
registry_by_scheduler = {
    'slurm': 'SLURMJobSubmitter',
    'oar': 'OARJobSubmitter',
}

def get_job_submitter(host: Optional[str]=None,
                      scheduler:Optional[str]=None,
                      **kwargs) -> JobSubmitter:
    """
    Factory function to get the JobSubmitter instance.

    Based on the input args `host` or `scheduler`, if an implemented JobSubmitter
    subclass can be located, return an instance of that subclass.
    If not, will use the base JobSubmitter class instance.
    
    Args:
        host (str, optional): Host machine name.

    Returns:
        JobSubmitter: An instance of the corresponding JobSubmitter subclass.
    """
    if host:
        host_name = host.lower()
        if host_name in registry_by_host.keys():
            module = importlib.import_module('NEDAS.job_submitters.'+host_name)
            JobSubmitterClass = getattr(module, registry_by_host[host_name])
            return JobSubmitterClass(**kwargs)

    if scheduler:
        scheduler_name = scheduler.lower()
        if scheduler_name in registry_by_scheduler.keys():
            module = importlib.import_module('NEDAS.job_submitters.'+scheduler_name)
            JobSubmitterClass = getattr(module, registry_by_scheduler[scheduler_name])
            return JobSubmitterClass(**kwargs)

    return JobSubmitter(**kwargs)



"""Utility funcs for rgps dataset"""
import numpy as np
from datetime import datetime, timedelta
from matplotlib.tri import Triangulation
from scipy.io import loadmat

def get_data_traj_pairs(file_name, d0_out, d1_out, dt_tol=2):
    """ Get rgps trajectory pairs x,y,t defined on rgps_proj, in km,day units"""

    ##several records, each contain pairs of points:
    ##x0,y0,t0 at time and x1,y1,t1 at time+dt

    pairs = []
    for stream in loadmat(file_name)['out'][0]:

        traj_x = [i[0] for i in stream['trajectories'][0][0]['x_map'][0]]
        traj_y = [i[0] for i in stream['trajectories'][0][0]['y_map'][0]]
        year = [i[0] for i in stream['trajectories'][0][0]['year'][0]]
        day = [i[0] for i in stream['trajectories'][0][0]['time'][0]]
        traj_d = [np.array([datetime(yr,1,1)+timedelta(d) for yr,d in zip(yr_,d_)])
                for yr_,d_ in zip(year, day)]

        xyd = ([],[],[],[],[],[])
        for x, y, d in zip(traj_x, traj_y, traj_d):
            if d[0] > d1_out or d[-1] < d0_out:
                continue  ##the target time is not in this traj
            dt0 = np.abs(d0_out - d)
            i0 = np.argmin(dt0)
            if dt0[i0] > dt_tol:
                continue
            dt1 = np.abs(d1_out - d)
            i1 = np.argmin(dt1)
            if dt1[i1] > dt_tol:
                continue

            _ = [i.append(j) for i,j in zip(xyd, [x[i0], y[i0], d[i0], x[i1], y[i1], d[i1]])]

        xyd = [np.array(i) for i in xyd]

        ##round time to days and select unique points in the pair
        rd = np.vstack([[np.round((d - d0_out).total_seconds()/3600) for d in dd] for dd in [xyd[2], xyd[5]]])
        rdu = np.unique(rd, axis=1)

        for rdu0, rdu1 in rdu.T:
            ind = np.logical_and(rd[0]==rdu0, rd[1]==rdu1)
            if len(xyd[0][ind]) < 3:  ##discard if too few points to form a mesh
                continue
            pairs.append([i[ind] for i in xyd])

    return pairs

def get_triangulation(x, y):
    """
    Get triangle indices, area, perimeter and mask
    for the RGPS trajectory points, x,y coordinates in km
    """

    tri = Triangulation(x, y)

    ##vertex coordinates
    xt, yt = [i[tri.triangles].T for i in (x, y)]

    ##side lengths
    tri_x = np.diff(np.vstack([xt, xt[0]]), axis=0)
    tri_y = np.diff(np.vstack([yt, yt[0]]), axis=0)
    tri_s = np.hypot(tri_x, tri_y)

    ##perimeter
    tri.p = np.sum(tri_s, axis=0)

    ##area
    s = tri.p / 2
    tri.a = np.sqrt(s * (s-tri_s[0]) * (s-tri_s[1]) * (s-tri_s[2]))

    ##mask off some triangles that don't belong to the mesh
    tri.mask = np.logical_or(tri.a>300, tri.p>50, tri.a/tri.p<1.5)

    return tri

def get_velocity(x0, y0, t0, x1, y1, t1):
    """get drift velocity in km/day"""
    d = ((t1 - t0) / timedelta(days=1)).astype(float)
    u, v = np.full(d.shape, np.nan), np.full(d.shape, np.nan)
    ind = (d>0)
    u[ind] = (x1[ind] - x0[ind]) / d[ind]
    v[ind] = (y1[ind] - y0[ind]) / d[ind]
    return u, v

def get_velocity_gradients(x, y, u, v):
    tri = get_triangulation(x, y)
    xt, yt, ut, vt = [i[tri.triangles].T for i in (x, y, u, v)]

    ux, uy, vx, vy = 0, 0, 0, 0
    for i0, i1 in zip([1, 2, 0], [0, 1, 2]):
        ux += (ut[i0] + ut[i1]) * (yt[i0] - yt[i1])
        uy -= (ut[i0] + ut[i1]) * (xt[i0] - xt[i1])
        vx += (vt[i0] + vt[i1]) * (yt[i0] - yt[i1])
        vy -= (vt[i0] + vt[i1]) * (xt[i0] - xt[i1])

    ux, uy, vx, vy = [i / (2 * tri.a) for i in (ux, uy, vx, vy)]
    return ux, uy, vx, vy

def get_deform_shear(x0, y0, t0, x1, y1, t1):
    u, v = get_velocity(x0, y0, t0, x1, y1, t1)
    ux, uy, vx, vy = get_velocity_gradients(x0, y0, u, v)
    return np.hypot(ux - vy, uy + vx)

def get_deform_div(x0, y0, t0, x1, y1, t1):
    u, v = get_velocity(x0, y0, t0, x1, y1, t1)
    ux, uy, vx, vy = get_velocity_gradients(x0, y0, u, v)
    return ux + vy

def get_deform_vort(x0, y0, t0, x1, y1, t1):
    u, v = get_velocity(x0, y0, t0, x1, y1, t1)
    ux, uy, vx, vy = get_velocity_gradients(x0, y0, u, v)
    return vx - uy



import numbers
import numpy as np
from datetime import datetime, timezone

def units_convert(units_from: str, units_to: str, var: np.ndarray) -> np.ndarray:
    """
    Convert units for a given variable.

    Args:
        units_from (str): Source units for the input variable
        units_to (str): Target units to convert to
        var (np.ndarray): The input variable

    Returns:
        np.ndarray: Variable with converted units
    """
    # if input units are numerics just apply the scaling factor
    if isinstance(units_to, numbers.Number) and isinstance(units_from, numbers.Number):
        return var * units_to / units_from

    # Define unit groups with a common base unit
    unit_groups = {
        "speed": {
            "base": "m/s",
            "to_base": {
                "km/h": lambda x: x / 3.6,
                "km/day": lambda x: x / 86.4,
                "cm/s": lambda x: x / 100.,
            },
            "from_base": {
                "km/h": lambda x: x * 3.6,
                "km/day": lambda x: x * 86.4,
                "cm/s": lambda x: x * 100.,
            },
        },
        "length": {
            "base": "m",
            "to_base": {
                "mm": lambda x: x / 1000.,
                "cm": lambda x: x / 100.,
                "dm": lambda x: x / 10.,
                "km": lambda x: x * 1000.,
            },
            "from_base": {
                "mm": lambda x: x * 1000.,
                "cm": lambda x: x * 100.,
                "dm": lambda x: x * 10.,
                "km": lambda x: x / 1000.,
            },
        },
        "time": {
            "base": "s",
            "to_base": {
                "min": lambda x: x * 60.,
                "h": lambda x: x * 3600.,
                "day": lambda x: x * 86400.,
            },
            "from_base": {
                "min": lambda x: x / 60.,
                "h": lambda x: x / 3600.,
                "day": lambda x: x / 86400.,
            },
        },
        "weight": {
            "base": "g",
            "to_base": {
                "kg": lambda x: x * 1000.,
            },
            "from_base": {
                "kg": lambda x: x / 1000.,
            },
        },
        "temperature": {
            "base": "K",
            "to_base": {
                "C": lambda x: x + 273.15,
                "F": lambda x: (x - 32) * 5 / 9 + 273.15,
            },
            "from_base": {
                "C": lambda x: x - 273.15,
                "F": lambda x: (x - 273.15) * 9 / 5 + 32,
            },
        },
        "pressure": {
            "base": "Pa",
            "to_base": {
                "hPa": lambda x: x * 100,
                "bar": lambda x: x * 100000.,
                "mbar": lambda x: x * 100.,
            },
            "from_base": {
                "hPa": lambda x: x / 100.,
                "bar": lambda x: x / 100000., 
                "mbar": lambda x: x / 100.,
            },
        },
        "energy_flux": {
            "base": "W/m2",
            "to_base": {
                "J/m2/d": lambda x: x / 86400.,
            },
            "from_base": {
                "J/m2/d": lambda x: x * 86400.,
            },
        },
        # Add other groups here...
    }

    if units_to == units_from:
        return var

    # Find the group containing the units
    for group, definitions in unit_groups.items():
        base = definitions["base"]
        to_base = definitions["to_base"]
        from_base = definitions["from_base"]

        # Check if both units exist in this group
        if units_from == base:
            if units_to in from_base:
                return from_base[units_to](var)
        elif units_to == base:
            if units_from in to_base:
                return to_base[units_from](var)
        elif units_from in to_base and units_to in from_base:
            # Convert to base, then from base to target
            var_base = to_base[units_from](var)
            return from_base[units_to](var_base)    

    raise ValueError(f"Conversion of unit from '{units_from}' to '{units_to}' not supported.")

##binary file io type conversion
type_convert = {'double':np.float64, 'float':np.float32, 'int':np.int32}
type_dic = {'double':'d', '8':'d', 'single':'f', 'float':'f', '4':'f', 'int':'i'}
type_size = {'double':8, 'float':4, 'int':4}

from pyproj import Proj
def proj2dict(proj: Proj) -> dict:
    """
    Convert map projection name in pyproj.Proj to a dictionary of human-readable parameters

    Args:
        proj (pyproj.Proj): Map projection object.

    Returns:
        dict: A dictionary of projection parameters, such as :code:`name`, :code:`lat_0`, :code:`lon_0`, etc.

    """
    proj_names = {
        "stere": "stereographic",
        "merc": "mercator",
        "lcc": "lambert_conformal_conic",
        "utm": "transverse_mercator",
        "aea": "albers_conical_equal_area",
        "eqc": "equirectangular",
        }
    params = {
        "proj":  {'name':"projection", 'type':str},
        "datum": {'name':"datum", 'type':str},
        "R":     {'name':"earth_radius", 'type':float},
        "lat_0": {'name':"latitude_of_projection_origin", 'type':float},
        "lon_0": {'name':"longitude_of_projection_origin", 'type':float},
        "lat_ts": {'name':"standard_parallel", 'type':float},
        "lat_1": {'name':"first_standard_parallel", 'type':float},
        "lat_2": {'name':"second_standard_parallel", 'type':float},
        "x_0": {'name':"false_easting", 'type':float},
        "y_0": {'name':"false_northing", 'type':float},
        "a": {'name':"semi_major_axis", 'type':float},
        "b": {'name':"semi_minor_axis", 'type':float},
        "k": {'name':"scale_factor_at_projection_origin", 'type':float},
    }
    proj_params = {}
    if not hasattr(proj, 'definition'):
        return {'projection':'customized'}
    for entry in proj.definition.split():
        if '=' in entry:
            key, value = entry.split('=', 1)
            if key in params:
                value = params[key]['type'](value)
        else:
            key, value = entry, True
        if key in params:
            if key == 'proj' and value in proj_names:
                value = proj_names[value]
            proj_params[params[key]['name']] = value
    return proj_params

from datetime import datetime, timedelta

dt1h = timedelta(hours=1)
ref_time = datetime(1900, 1, 1, tzinfo=timezone.utc)

def t2h(t: datetime) -> float:
    """
    Convert datetime object to hours since 1900-1-1 00:00
    """
    return (t - ref_time)/timedelta(hours=1)

def h2t(h: float) -> datetime:
    """
    Convert hours since 1900-1-1 00:00 to datetime object
    """
    return ref_time + timedelta(hours=1) * h

def t2s(t: datetime) -> str:
    """
    Convert datetime object to a time string :code:`'ccyymmddHHMM'`
    """
    return t.strftime('%Y%m%d%H%M')

def s2t(s: str) -> datetime:
    """
    Convert a time string :code:`'ccyymmddHHMM'` to a datetime object
    """
    return datetime.strptime(s, '%Y%m%d%H%M')

def seconds_to_timestr(seconds: int) -> str:
    """
    Convert from seconds to time duration string 'HH:MM:SS'
    """
    hours = seconds // 3600
    minutes = (seconds % 3600) // 60
    seconds = seconds % 60
    return f"{hours:02}:{minutes:02}:{seconds:02}"

def ensure_list(v) -> list:
    """
    If the input :code:`v` is a list, return itself; if not, return :code:`[v]`.
    """
    if isinstance(v, list):
        return v
    return [v]




import unittest
from NEDAS.config import Config
from NEDAS.schemes.get_analysis_scheme import get_analysis_scheme

class TestAnalysisScheme(unittest.TestCase):
    def test_raise_exception_when_not_implemented(self):        
        with self.assertRaises(NotImplementedError):
            c = Config(analysis_scheme='foo')
            get_analysis_scheme(c)

if __name__ == '__main__':
    unittest.main()



import numpy as np
import os
import sys
from NEDAS.models.qg import Model
from NEDAS.models.qg.util import read_data_bin, spec2grid
from NEDAS.utils.conversion import t2s, s2t, dt1h
from NEDAS.utils.netcdf_lib import nc_write_var

def main():
    model_name = sys.argv[1] ##'qg'
    nens = int(sys.argv[2]) ##20
    tstr = sys.argv[3]  ##'202301021200'

    mem_list = np.arange(nens)
    model = Model()

    kmax = 127
    nx = 2*(kmax+1)
    nz = 2
    path = '/cluster/work/users/yingyue/'+model_name+f'.n{nens}'+'/cycle/'+tstr+'/'+model_name
    t = s2t(tstr)

    vnames = ['temperature', 'streamfunc', 'vorticity']

    for m in mem_list:
        print(m)
        for n in [0, 1]:
            tout = t + n*12*dt1h
            for name in vnames:
                field = np.zeros((nz, nx, nx))
                for k in range(nz):
                    field[k,...] = model.read_var(name=name, path=path, member=m, time=tout, k=k)
                nc_write_var(path+'/ens.nc', {'t':None, 'm':None, 'z':nz, 'y':nx, 'x':nx}, name, field, recno={'t':n, 'm':m})

if __name__ == '__main__':
    main()




import os
from NEDAS.config import Config
from NEDAS.utils.conversion import dt1h

def main():
    c = Config(parse_args=True)

    model = c.models['vort2d']

    truth_dir = model.truth_dir
    os.system("mkdir -p "+truth_dir)

    t = c.time_start
    while t < c.time_end:
        opts = {
            'path': truth_dir,
            'name': 'velocity',
            'is_vector': True,
            'time': t,
            }

        if t == c.time_start:
            state = model.generate_initial_condition()
            print(f"generating initial condition {model.filename(**opts)}")
            model.write_var(state, **opts)

        next_t = t + c.cycle_period * dt1h
        print(f"running model, saving output {model.filename(**{**opts, 'time':next_t})}")
        model.run(path=truth_dir, time=t, forecast_period=c.cycle_period)

        t = next_t

    print("done.")

if __name__ == '__main__':
    main()



import os
import glob
import numpy as np
from datetime import datetime, timedelta, timezone
from NEDAS.utils.conversion import dt1h
from NEDAS.datasets import Dataset

class ArgoObs(Dataset):

    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.variables = {
            'ocean_temp':{'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'K'},
            'ocean_saln':{'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'psu'},
            }
        self.obs_operator = {}

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        path = kwargs['path']
        time = kwargs['time']
        obs_window_min = kwargs['obs_window_min']
        obs_window_max = kwargs['obs_window_max']

        if time is None:
            tstr = '????????'
            search = os.path.join(path, tstr+'_?_prof.nc')
            file_list = glob.glob(search)

        else:
            if obs_window_min is not None and obs_window_max is not None:
                d_range = np.arange(obs_window_min, obs_window_max)
            else:
                d_range = [0]

            file_list = []
            for d in d_range:
                t = time + d * dt1h
                tstr = t.strftime('%Y%m%d')
                search = os.path.join(path, tstr+'_?_prof.nc')
                for result in glob.glob(search):
                    if result not in file_list:
                        file_list.append(result)

        assert len(file_list)>0, 'no matching files found: '+search
        return file_list

    def read_obs(self, **kwargs):
        """
        read raw data file, quality check, thinning according to model grid, then output to obs_seq
        kwargs:
        -grid: model 2D grid [ny, nx]
        -mask: masked or not (for land or other reason) bool[ny, nx]
        -z: depth (-meters) defined on grid [nz, ny, nx], if None not available
        -name, time, etc.
        return: obs_seq dict with lists of obs,x,y,z,time
        """
        kwargs = super().parse_kwargs(**kwargs)
        name = kwargs['name']
        
        ##target grid for obs_seq
        grid = kwargs['grid']
        mask = kwargs['mask']

        model = kwargs['model']
        assert model is not None, 'read_obs: ERROR: model is required for argo dataset'
        model.grid.set_destination_grid(grid)
      
        obs_seq = {'obs':[],
                   't':[], 'z':[], 'y':[], 'x':[],
                   'err_std':[],
                   'profile_id':[], 'level_id':[] }

        for fname in self.filename(**kwargs):

            ##read the profiles from nc file
            f = Dataset(fname)

            nprof = f.dimensions['N_PROF'].size
            nlev = f.dimensions['N_LEVELS'].size

            lat = f['LATITUDE'][0:nprof].data
            lon = f['LONGITUDE'][0:nprof].data
            x, y = grid.proj(lon, lat)  ##coordinates in grid

            juld = f['JULD'][0:nprof].data
            juld_qc = f['JULD_QC'][0:nprof].data
            pos_qc = f['POSITION_QC'][0:nprof].data

            ##find z coordinate in meters, negative relative to surface
            if 'PRES' in f.variables:
                z = 0. - f['PRES'][0:nprof, 0:nlev].data * 1e4 / self.ONEM ##decibar to meter
                z_qc = f['PRES_QC'][0:nprof, 0:nlev].data
            elif 'DEPH' in f.variables:
                z = 0. - f['DEPH'][0:nprof, 0:nlev].data
                z_qc = f['DEPH_QC'][0:nprof, 0:nlev].data

            ##model z (zm) at profile location on grid
            if model.z is not None:
                nz = model.z.shape[0]
                zm = np.zeros((nprof, nz))
                for k in range(nz):
                    ztmp = model.grid.convert(model.z[k, ...])
                    zm[:, k] = grid.interp(ztmp, x, y)

            ##observed variable
            if name == 'ocean_temp' and 'TEMP' in f.variables:
                obs = f['TEMP'][0:nprof, 0:nlev].data
                obs_qc = f['TEMP_QC'][0:nprof, 0:nlev].data
            elif name == 'ocean_saln' and 'PSAL' in f.variables:
                obs = f['PSAL'][0:nprof, 0:nlev].data
                obs_qc = f['PSAL_QC'][0:nprof, 0:nlev].data
            else:
                continue

            ##quality check flags
            flag1 = np.ones(nprof)
            flag2 = np.ones((nprof, nlev))

            for p in range(nprof):
                if juld_qc[p] not in (b'1', b'2'):
                    flag1[p] = 0
                if pos_qc[p] not in (b'1', b'2'):
                    flag1[p] = 0

            ##check inside grid
            for p in range(nprof):
                if x[p]<grid.xmin or x[p]>grid.xmax or y[p]<grid.ymin or y[p]>grid.ymax:
                    flag1[p] = 0

            for p in range(nprof):
                for l in range(nlev):
                    if z_qc[p,l] not in (b'1', b'2'):
                        flag2[p,l] = 0
                    if obs_qc[p,l] not in (b'1',):
                        flag2[p,l] = 0
                if np.sum(flag2[p,:]) == 0:
                    flag1[p] = 0

            ##check if location is unmasked (wet)
            mask_prof = grid.interp(mask.astype(int), x, y)
            for p in range(nprof):
                if flag1[p] == 0:
                    continue
                ##check if profile location is masked in model grid
                if mask_prof[p] > 0:
                    flag1[p] = 0
                    continue
                ##check if z level is deeper than model depth at profile location
                if model.z is not None:
                    for l in range(nlev):
                        if z[p,l] < np.min(zm[p,:]):
                            flag2[p,l] = 0
                    if np.sum(flag2[p,:]) == 0:
                        flag1[p] = 0

            ##check physical value
            for p in range(nprof):
                if flag1[p] == 0:
                    continue
                for l in range(nlev):
                    if flag2[p,l] == 0:
                        continue
                    if name == 'ocean_temp':
                        if obs[p,l] < self.TEMP_MIN or obs[p,l] > self.TEMP_MAX:
                            flag1[p] = 0
                            flag2[p,:] = 0
                    if name == 'ocean_saln':
                        if obs[p,l] < self.SALN_MIN or obs[p,l] > self.SALN_MAX:
                            flag1[p] = 0
                            flag2[p,:] = 0

            ##TODO: thinning in horizontal

            ##thinning in vertical
            if self.NUM_OBS_PER_LAYER is not None and model.z is not None:
                nz = model.z.shape[0]
                for p in range(nprof):
                    if flag1[p] == 0:
                        continue
                    ##target z levels to bin profile obs
                    zt = np.interp(np.arange(0, nz, 1/self.NUM_OBS_PER_LAYER),
                                   np.arange(1, nz+1, 1), zm[p, :])
                    zt = np.sort(zt)
                    zt_ind = []
                    for l in range(nlev):
                        if flag2[p,l] == 0:
                            continue
                        ind = np.searchsorted(zt, z[p, l])
                        if ind not in zt_ind:
                            zt_ind.append(ind)
                        else:
                            flag2[p,l] = 0

            ##check convective stability is not implemented here
            ##  potential_density(temp, saln), find vertical increment from level k to k+1
            ##  if density difference rho_inc < DENS_DIFF_MIN = -0.02
            ##  then discard this profile

            ##output data to the obs_seq
            for p in range(nprof):
                if flag1[p] == 0:
                    continue

                ##time of the profile given by julian day
                t_prof = datetime(1950,1,1,tzinfo=timezone.utc) + juld[p]*timedelta(days=1)

                for l in range(nlev):
                    if flag2[p,l] == 0:
                        continue
                    obs_seq['obs'].append(obs[p,l])
                    obs_seq['t'].append(t_prof)
                    obs_seq['z'].append(z[p,l])
                    obs_seq['y'].append(y[p])
                    obs_seq['x'].append(x[p])
                    obs_seq['err_std'].append(np.sqrt(self.OBS_ERR_VAR[name]))
                    obs_seq['profile_id'].append(p)
                    obs_seq['level_id'].append(l)

            f.close()

        for key in obs_seq.keys():
            obs_seq[key] = np.array(obs_seq[key])

        return obs_seq

    def random_network(self, model, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        raise NotImplementedError
        obs_seq = {}
        return obs_seq



import copy
from abc import abstractmethod
import numpy as np
from NEDAS.utils.parallel import distribute_tasks
from NEDAS.utils.progress import progress_bar
from NEDAS.assim_tools.assimilators.base import Assimilator

class BatchAssimilator(Assimilator):
    def init_partitions(self, c):
        """
        Generate spatial partitioning of the domain
        partitions: dict[par_id, tuple(istart, iend, di, jstart, jend, dj)]
        for each partition indexed by par_id, the tuple contains indices for slicing the domain
        Using regular slicing is more efficient than fancy indexing (used in irregular grid)
        """
        if len(c.grid.x.shape) == 2:
            ny, nx = c.grid.x.shape

            ##divide into square tiles with nx_tile grid points in each direction
            ##the workload on each tile is uneven since there are masked points
            ##so we divide into 3*nproc tiles so that they can be distributed
            ##according to their load (number of unmasked points)
            ntile = c.nproc_mem * 3
            nx_tile = np.maximum(int(np.round(np.sqrt(nx * ny / ntile))), 1)

            ##a list of (istart, iend, di, jstart, jend, dj) for tiles
            ##note: we have 3*nproc entries in the list
            partitions = [(i, np.minimum(i+nx_tile, nx), 1,   ##istart, iend, di
                           j, np.minimum(j+nx_tile, ny), 1)   ##jstart, jend, dj
                          for j in np.arange(0, ny, nx_tile)
                          for i in np.arange(0, nx, nx_tile) ]

        else:
            npoints = c.grid.x.size
            ##divide the domain into sqaure tiles, similar to regular_grid case, but collect
            ##the grid points inside each tile and return the indices
            ntile = c.nproc_mem * 3

            if c.grid.Ly==0:
                ##for 1D grid, just divide into equal sections, no y dimension
                Dx = c.grid.Lx / ntile
                partitions = [np.where(np.logical_and(c.grid.x>=x, c.grid.x<x+Dx))[0]
                              for x in np.arange(c.grid.xmin, c.grid.xmax, Dx)]

            else:
                ##for 2D grid, find number of tiles in each direction according to aspect ratio
                ntile_y = max(int(np.sqrt(ntile * c.grid.Ly / c.grid.Lx)), 1)
                ntile_x = max(ntile // ntile_y, 1)
                Dx = c.grid.Lx / ntile_x
                Dy = c.grid.Ly / ntile_y
                partitions = [np.where(np.logical_and(np.logical_and(c.grid.x>=x, c.grid.x<x+Dx),
                                                      np.logical_and(c.grid.y>=y, c.grid.y<y+Dy)))
                              for y in np.arange(c.grid.ymin, c.grid.ymax, Dy)
                              for x in np.arange(c.grid.xmin, c.grid.xmax, Dx)]

        return partitions

    def assign_obs(self, c, state, obs):
        """
        Assign the observation sequence to each partition par_id

        Returns:
        - obs_inds: dict[obs_rec_id, dict[par_id, inds]]
        where inds is np.array with indices in the full obs_seq, for the subset of obs
        that belongs to partition par_id
        """
        ##each pid_rec has a subset of obs_rec_list
        obs_inds_pid = {}
        for obs_rec_id in obs.obs_rec_list[c.pid_rec]:
            ##screen horizontally for obs inside hroi of each partition
            obs_inds_pid[obs_rec_id] = self.assign_obs_to_tiles(c, state, obs, obs_rec_id)

        ##now each pid_rec has figured out obs_inds for its own list of obs_rec_ids, we
        ##gather all obs_rec_id from different pid_rec to form the complete obs_inds dict
        obs_inds = {}
        for entry in c.comm_rec.allgather(obs_inds_pid):
            for obs_rec_id, data in entry.items():
                obs_inds[obs_rec_id] = data

        return obs_inds

    def assign_obs_to_tiles(self, c, state, obs, obs_rec_id):
        hroi = obs.info['records'][obs_rec_id]['hroi']

        xo = np.array(obs.obs_seq[obs_rec_id]['x'])  ##obs x,y
        yo = np.array(obs.obs_seq[obs_rec_id]['y'])

        ##loop over partitions with par_id
        obs_inds = {}
        for par_id in range(len(state.partitions)):
            ##find bounding box for this partition
            if len(c.grid.x.shape)==2:
                ist,ied,di,jst,jed,dj = state.partitions[par_id]
                xmin, xmax, ymin, ymax = c.grid.x[0,ist], c.grid.x[0,ied-1], c.grid.y[jst,0], c.grid.y[jed-1,0]
            else:
                inds = state.partitions[par_id]
                x = c.grid.x[inds]
                y = c.grid.y[inds]
                xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
            Dx = 0.5 * (xmax - xmin)
            Dy = 0.5 * (ymax - ymin)
            xc = xmin + Dx
            yc = ymin + Dy

            ##observations within the bounding box + halo region of width hroi will be assigned to
            ##this partition. Although this will include some observations near the corner that are
            ##not within hroi of any grid points, this is favorable for the efficiency in finding subset
            obs_inds[par_id] = np.where(np.logical_and(c.grid.distance(xc, xo, yc, yc, p=1) <= Dx+hroi,
                                                       c.grid.distance(xc, xc, yc, yo, p=1) <= Dy+hroi))[0]

        return obs_inds

    def distribute_partitions(self, c, state, obs):
        par_list_full = np.arange(len(state.partitions))

        ##distribute the list of par_id according to workload to each pid
        ##number of unmasked grid points in each tile
        if len(c.grid.x.shape) == 2:
            nlpts_loc = np.array([np.sum((~c.grid.mask[jst:jed:dj, ist:ied:di]).astype(int))
                                 for ist,ied,di,jst,jed,dj in state.partitions] )
        else:
            nlpts_loc = np.array([np.sum((~c.grid.mask[inds]).astype(int))
                                 for inds in state.partitions] )

        ##number of observations within the hroi of each tile, at loc,
        ##sum over the len of obs_inds for obs_rec_id over all obs_rec_ids
        nlobs_loc = np.array([np.sum([len(obs.obs_inds[r][p])
                                      for r in obs.info['records'].keys()])
                              for p in par_list_full] )

        workload = np.maximum(nlpts_loc, 1) * np.maximum(nlobs_loc, 1)
        par_list = distribute_tasks(c.comm_mem, par_list_full, workload)

        return par_list

    def assimilation_algorithm(self, c, state, obs):
        """
        batch assimilation solves the matrix version EnKF analysis for each local state,
        the local states in each partition are processed in parallel
        """
        state.state_post = copy.deepcopy(state.state_prior)
        obs.lobs_post = copy.deepcopy(obs.lobs_prior)

        ##pid with the most obs in its task list with show progress message
        obs_count = np.array([np.sum([len(obs.obs_inds[r][p])
                                      for r in obs.info['records'].keys()
                                      for p in lst])
                              for lst in state.par_list.values()])
        c.pid_show = np.argsort(obs_count)[-1]

        ##count number of tasks
        ntask = 0
        for par_id in state.par_list[c.pid_mem]:
            if len(c.grid.x.shape)==2:
                ist,ied,di,jst,jed,dj = state.partitions[par_id]
                msk = c.grid.mask[jst:jed:dj, ist:ied:di]
            else:
                inds = state.partitions[par_id]
                msk = c.grid.mask[inds]
            for loc_id in range(np.sum((~msk).astype(int))):
                ntask += 1

        ##now the actual work starts, loop through partitions stored on pid_mem
        c.print_1p('>>> assimilate in batch mode:\n')
        task = 0
        for par_id in state.par_list[c.pid_mem]:
            state_data = state.pack_local_state_data(c, par_id, state.state_prior, state.z_state)
            nloc = state_data['state_prior'].shape[-1]
            ##skip forward if the partition is empty
            if nloc == 0:
                continue

            obs_data = obs.pack_local_obs_data(c, state, par_id, obs.lobs, obs.lobs_prior)
            nlobs = obs_data['x'].size
            ##if there is no obs to assimilate, update progress message and skip that partition
            if nlobs == 0:
                task += nloc
                if c.debug:
                    print(f"PID {c.pid:4} processed partition {par_id:7} (which is empty)", flush=True)
                else:
                    c.print_1p(progress_bar(task-1, ntask))
                continue

            ##loop through the unmasked grid points in the partition
            for loc_id in range(nloc):
                ##state variable metadata for this location
                state_x = state_data['x'][loc_id]
                state_y = state_data['y'][loc_id]

                ##filter out obs outside the hroi in each direction first (using L1 norm to speed up)
                obs_rec_id = obs_data['obs_rec_id']
                hroi = obs_data['hroi'][obs_rec_id]
                hdist = c.grid.distance(state_x, obs_data['x'], state_y, obs_data['y'], p=1)
                ind = np.where(hdist<=hroi)[0]

                ##TODO: filter out nan in obs_prior
                #obs_data['obs_prior']

                ##compute horizontal localization factor (using L2 norm for distance)
                obs_rec_id = obs_data['obs_rec_id'][ind]
                hroi = obs_data['hroi'][obs_rec_id]
                hdist = c.grid.distance(state_x, obs_data['x'][ind], state_y, obs_data['y'][ind], p=2)
                hlfactor = c.localization_funcs['horizontal'](hdist, hroi)
                ind1 = np.where(hlfactor>0)[0]
                ind = ind[ind1]
                hlfactor = hlfactor[ind1]

                if len(ind1) == 0:
                    if c.debug:
                        print(f"PID {c.pid:4} processed partition {par_id:7} grid point {loc_id} (all local obs outside hroi)", flush=True)
                    else:
                        c.print_1p(progress_bar(task, ntask))
                    continue ##if all obs has no impact on state, just skip to next location

                self.local_analysis(c, loc_id, ind, hlfactor, state_data, obs_data)

                ##add progress message
                if c.debug:
                    print(f"PID {c.pid:4} processed partition {par_id:7} grid point {loc_id}", flush=True)
                else:
                    c.print_1p(progress_bar(task, ntask))
                task += 1

            state.unpack_local_state_data(c, par_id, state.state_post, state_data)
            obs.unpack_local_obs_data(c, state, par_id, obs.lobs, obs.lobs_post, obs_data)
        c.print_1p(' done.\n')

    @abstractmethod
    def local_analysis(self, c, loc_id, ind, hlfactor, state_data, obs_data):
        """Local analysis scheme for each model state variable (grid point)
        to be implemented by derived classes"""
        pass


import numpy as np
from NEDAS.models.lorenz96 import Lorenz96Model
from NEDAS.datasets import Dataset

class Lorenz96Obs(Dataset):

    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.variables = {
            'state': {'dtype':'float', 'is_vector':False, 'z_units':'*', 'units':'*'},
            }

    def random_network(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        assert isinstance(kwargs['model'], Lorenz96Model), 'random_network: ERROR: model must be an instance of models.lorenz96.Model'
        grid = kwargs['grid']

        nobs = kwargs['nobs']
        obs_x = np.random.uniform(grid.xmin, grid.xmax, nobs)

        obs_seq = {'obs': np.full(nobs, np.nan),
                't': np.full(nobs, kwargs['time']),
                'z': np.zeros(nobs),
                'y': np.zeros(nobs),
                'x': obs_x,
                'err_std': np.ones(nobs) * kwargs['err']['std'],
                }

        return obs_seq

    def read_obs(self):
        raise NotImplementedError('Only synthetic obs will be used.')